from enum import Enum
import copy
import numpy as np
from numpy.linalg import norm
from scipy.spatial.transform import Rotation

class ExitStatus(Enum):
    """ Exit status values indicate the reason for simulation termination. """
    COMPLETE     = 'Success: End reached.'
    TIMEOUT      = 'Timeout: Simulation end time reached.'
    INF_VALUE    = 'Failure: Your controller returned inf motor speeds.'
    NAN_VALUE    = 'Failure: Your controller returned nan motor speeds.'
    OVER_SPEED   = 'Failure: Your quadrotor is out of control; it is going faster than 100 m/s. The Guinness World Speed Record is 73 m/s.'
    OVER_SPIN    = 'Failure: Your quadrotor is out of control; it is spinning faster than 100 rad/s. The onboard IMU can only measure up to 52 rad/s (3000 deg/s).'
    FLY_AWAY     = 'Failure: Your quadrotor is out of control; it flew away with a position error greater than 20 meters.'
    COLLISION    = 'Failure: Your quadrotor collided with an object.'

def simulate(world, initial_state, vehicle, controller, trajectory, wind_profile, imu, mocap, estimator, t_final, t_step, safety_margin, use_mocap, terminate=None):
    """
    Perform a vehicle simulation and return the numerical results.

    Inputs:
        world, a class representing the world it is flying in, including objects and world bounds. 
        initial_state, a dict defining the vehicle initial conditions with appropriate keys
        vehicle, Vehicle object containing the dynamics
        controller, Controller object containing the controller
        trajectory, Trajectory object containing the trajectory to follow
        wind_profile, Wind Profile object containing the wind generator. 
        t_final, maximum duration of simulation, s
        t_step, the time between each step in the simulator, s
        safety_margin, the radius of the ball surrounding the vehicle position to determine if a collision occurs
        imu, IMU object that generates accelerometer and gyroscope readings from the vehicle state
        terminate, None, False, or a function of time and state that returns
            ExitStatus. If None (default), terminate when hover is reached at
            the location of trajectory with t=inf. If False, never terminate
            before timeout or error. If a function, terminate when returns not
            None.
        mocap, a MotionCapture object that provides noisy measurements of pose and twist with artifacts. 
        use_mocap, a boolean to determine in noisy measurements from mocap should be used for quadrotor control
        estimator, an estimator object that provides estimates of a portion or all of the vehicle state.

    Outputs:
        time, seconds, shape=(N,)
        state, a dict describing the state history with keys
            x, position, m, shape=(N,3)
            v, linear velocity, m/s, shape=(N,3)
            q, quaternion [i,j,k,w], shape=(N,4)
            w, angular velocity, rad/s, shape=(N,3)
            rotor_speeds, motor speeds, rad/s, shape=(N,n) where n is the number of rotors
            wind, wind velocity, m/s, shape=(N,3)
        control, a dict describing the command input history with keys
            cmd_motor_speeds, motor speeds, rad/s, shape=(N,4)
            cmd_q, commanded orientation (not used by simulator), quaternion [i,j,k,w], shape=(N,4)
            cmd_w, commanded angular velocity (not used by simulator), rad/s, shape=(N,3)
        flat, a dict describing the desired flat outputs from the trajectory with keys
            x,        position, m
            x_dot,    velocity, m/s
            x_ddot,   acceleration, m/s**2
            x_dddot,  jerk, m/s**3
            x_ddddot, snap, m/s**4
            yaw,      yaw angle, rad
            yaw_dot,  yaw rate, rad/s
        imu_measurements, a dict containing the biased and noisy measurements from an accelerometer and gyroscope
            accel,  accelerometer, m/s**2
            gyro,   gyroscope, rad/s
        imu_gt, a dict containing the ground truth (no noise, no bias) measurements from an accelerometer and gyroscope
            accel,  accelerometer, m/s**2
            gyro,   gyroscope, rad/s
        mocap_measurements, a dict containing noisy measurements of pose and twist for the vehicle. 
            x, position (inertial)
            v, velocity (inertial)
            q, orientation of body w.r.t. inertial frame.
            w, body rates in the body frame. 
        exit_status, an ExitStatus enum indicating the reason for termination.
    """

    # Coerce entries of initial state into numpy arrays, if they are not already.
    initial_state = {k: np.array(v) for k, v in initial_state.items()}

    if terminate is None:    # Default exit. Terminate at final position of trajectory.
        normal_exit = traj_end_exit(initial_state, trajectory, using_vio = False)
    elif terminate is False: # Never exit before timeout.
        normal_exit = lambda t, s: None
    else:                    # Custom exit.
        normal_exit = terminate

    time    = [0]
    state   = [copy.deepcopy(initial_state)]
    state[0]['wind'] = wind_profile.update(0, state[0]['x'])   # TODO: move this line elsewhere so that other objects that don't have wind as a state can work here. 
    imu_measurements = []
    mocap_measurements = []
    imu_gt = []
    state_estimate = []
    flat    = [trajectory.update(time[-1])]
    mocap_measurements.append(mocap.measurement(state[-1], with_noise=True, with_artifacts=False))
    if use_mocap:
        # In this case the controller will use the motion capture estimate of the pose and twist for control. 
        control = [controller.update(time[-1], mocap_measurements[-1], flat[-1])]
    else:
        control = [controller.update(time[-1], state[-1], flat[-1])]
    state_dot =  vehicle.statedot(state[0], control[0], t_step)
    imu_measurements.append(imu.measurement(state[-1], state_dot, with_noise=True))
    imu_gt.append(imu.measurement(state[-1], state_dot, with_noise=False))
    state_estimate.append(estimator.step(state[0], control[0], imu_measurements[0], mocap_measurements[0]))

    exit_status = None

    while True:
        exit_status = exit_status or safety_exit(world, safety_margin, state[-1], flat[-1], control[-1])
        exit_status = exit_status or normal_exit(time[-1], state[-1])
        exit_status = exit_status or time_exit(time[-1], t_final)
        if exit_status:
            break
        time.append(time[-1] + t_step)
        state[-1]['wind'] = wind_profile.update(time[-1], state[-1]['x'])
        state.append(vehicle.step(state[-1], control[-1], t_step))
        flat.append(trajectory.update(time[-1]))
        mocap_measurements.append(mocap.measurement(state[-1], with_noise=True, with_artifacts=mocap.with_artifacts))
        state_estimate.append(estimator.step(state[-1], control[-1], imu_measurements[-1], mocap_measurements[-1]))
        if use_mocap:
            control.append(controller.update(time[-1], mocap_measurements[-1], flat[-1]))
        else:
            control.append(controller.update(time[-1], state[-1], flat[-1]))
        state_dot = vehicle.statedot(state[-1], control[-1], t_step)
        imu_measurements.append(imu.measurement(state[-1], state_dot, with_noise=True))
        imu_gt.append(imu.measurement(state[-1], state_dot, with_noise=False))

    time    = np.array(time, dtype=float)    
    state   = merge_dicts(state)
    imu_measurements = merge_dicts(imu_measurements)
    imu_gt = merge_dicts(imu_gt)
    mocap_measurements = merge_dicts(mocap_measurements)
    control         = merge_dicts(control)
    flat            = merge_dicts(flat)
    state_estimate  = merge_dicts(state_estimate)

    return (time, state, control, flat, imu_measurements, imu_gt, mocap_measurements, state_estimate, exit_status)

def merge_dicts(dicts_in):
    """
    Concatenates contents of a list of N state dicts into a single dict by
    prepending a new dimension of size N. This is more convenient for plotting
    and analysis. Requires dicts to have consistent keys and have values that
    are numpy arrays.
    """
    dict_out = {}
    for k in dicts_in[0].keys():
        dict_out[k] = []
        for d in dicts_in:
            dict_out[k].append(d[k])
        dict_out[k] = np.array(dict_out[k])
    return dict_out


def traj_end_exit(initial_state, trajectory, using_vio = False):
    """
    Returns a exit function. The exit function returns an exit status message if
    the quadrotor is near hover at the end of the provided trajectory. If the
    initial state is already at the end of the trajectory, the simulation will
    run for at least one second before testing again.
    """

    xf = trajectory.update(np.inf)['x']
    yawf = trajectory.update(np.inf)['yaw']
    rotf = Rotation.from_rotvec(yawf * np.array([0, 0, 1])) # create rotation object that describes yaw
    if np.array_equal(initial_state['x'], xf):
        min_time = 1.0
    else:
        min_time = 0

    def exit_fn(time, state):
        cur_attitude = Rotation.from_quat(state['q'])
        err_attitude = rotf * cur_attitude.inv() # Rotation between current and final
        angle = norm(err_attitude.as_rotvec()) # angle in radians from vertical
        # Success is reaching near-zero speed with near-zero position error.
        if using_vio:
            # set larger threshold for VIO due to noisy measurements
            if time >= min_time and norm(state['x'] - xf) < 1 and norm(state['v']) <= 1 and angle <= 1:
                return ExitStatus.COMPLETE
        else:
            if time >= min_time and norm(state['x'] - xf) < 0.02 and norm(state['v']) <= 0.03 and angle <= 0.02:
                return ExitStatus.COMPLETE
        return None
    return exit_fn

def time_exit(time, t_final):
    """
    Return exit status if the time exceeds t_final, otherwise None.
    """
    if time >= t_final:
        return ExitStatus.TIMEOUT
    return None

def safety_exit(world, margin, state, flat, control):
    """
    Return exit status if any safety condition is violated, otherwise None.
    """
    if np.any(np.abs(state['v']) > 20):
        return ExitStatus.OVER_SPEED
    if np.any(np.abs(state['w']) > 100):
        return ExitStatus.OVER_SPIN

    if len(world.world.get('blocks', [])) > 0:
        # If a world has objects in it we need to check for collisions.  
        collision_pts = world.path_collisions(state['x'], margin)
        no_collision = collision_pts.size == 0
        if not no_collision:
            return ExitStatus.COLLISION
    return None


"""
Imports
"""
import numpy as np
import cvxopt
from scipy.linalg import block_diag

def cvxopt_solve_qp(P, q, G=None, h=None, A=None, b=None):
    """
    From https://scaron.info/blog/quadratic-programming-in-python.html . Infrastructure code for solving quadratic programs using CVXOPT. 
    The structure of the program is as follows: 

    min 0.5 xT P x + qT x
    s.t. Gx <= h
         Ax = b
    Inputs:
        P, numpy array, the quadratic term of the cost function
        q, numpy array, the linear term of the cost function
        G, numpy array, inequality constraint matrix
        h, numpy array, inequality constraint vector
        A, numpy array, equality constraint matrix
        b, numpy array, equality constraint vector
    Outputs:
        The optimal solution to the quadratic program
    """
    P = .5 * (P + P.T)  # make sure P is symmetric
    args = [cvxopt.matrix(P), cvxopt.matrix(q)]
    if G is not None:
        args.extend([cvxopt.matrix(G), cvxopt.matrix(h)])
    if A is not None:
        args.extend([cvxopt.matrix(A), cvxopt.matrix(b)])
    sol = cvxopt.solvers.qp(*args)
    if 'optimal' not in sol['status']:
        return None
    return np.array(sol['x']).reshape((P.shape[1],))

def H_fun(dt, k=7):
    """
    Computes the cost matrix for a single segment in a single dimension.
    *** Assumes that the decision variables c_i are e.g. x(t) = c_0 + c_1*t + c_2*t^2 + c_3*t^3 + c_4*t^4 + c_5*t^5 + .. + c_k*t^k
    Inputs:
        dt, scalar, the duration of the segment (t_(i+1) - t_i) 
        k, scalar, the order of the polynomial. 
    Outputs:
        H, numpy array, matrix containing the min snap cost function for that segment. Assumes the polynomial is at least order 5.
    """

    H = np.zeros((k+1,k+1))

    # TODO: implement automatic cost function based on the desired cost functional (snap or jerk) and the order of the polynomial, k. 

    seventh_order_cost = np.array([[576*dt, 1440*dt**2, 2880*dt**3, 5040*dt**4],
                                   [1440*dt**2, 4800*dt**3, 10800*dt**4, 20160*dt**5],
                                   [2880*dt**3, 10800*dt**4, 25920*dt**5, 50400*dt**6],
                                   [5040*dt**4, 20160*dt**5, 50400*dt**6, 100800*dt**7]])

    # Only take up to the (k+1) entries
    cost = seventh_order_cost[0:(k+1-4), 0:(k+1-4)]

    H[4:(k+1),4:(k+1)] = cost

    return H 

def get_1d_constraints(keyframes, delta_t, m, k=7, vmax=5, vstart=0, vend=0):
    """
    Computes the constraint matrices for the min snap problem. 
    *** Assumes that the decision variables c_i are e.g. o(t) = c_0 + c_1*t + c_2*t^2 + ... c_(k)*t^(k)

    We impose the following constraints FOR EACH SEGMENT m: 
        1) x_m(0) = keyframe[i]             # position at t = 0
        2) x_m(dt) = keyframe[i+1]          # position at t = dt
        3) v_m(0) = v_start                 # velocity at t = 0
        4) v_m(dt) = v_end                  # velocity at t = dt
        5) v_m(dt) = v_(m+1)(0)             # velocity continuity for interior segments
        6) a_m(dt) = a_(m+1)(0)             # acceleration continuity for interior segments
        7) j_m(dt) = j_(m+1)(0)             # jerk continuity for interior segments
        8) s_m(dt) = s_(m+1)(0)             # snap continuity for interior segments
        9) v_m(dt/2) <= vmax                # velocity constraint at midpoint of each segment

    For the first and last segment we impose: 
        1) a_0(0) = 0                       # acceleration at start of the trajectory is 0
        2) j_0(0) = 0                       # jerk at start of the trajectory is 0
        3) a_N(dt) = 0                      # acceleration at the end of the trajectory is 0
        4) j_N(dt) = 0                      # jerk at the end of the trajectory is 0

    Inputs:
        keyframes, numpy array, a list of m waypoints IN ONE DIMENSION (x,y,z, or yaw)
        delta_t, numpy array, the times between keyframes computed apriori. 
        m, int, the number of segments.
        k, int, the degree of the polynomial. 
        vmax, float, max speeds imposed at the midpoint of each segment. 
        vstart, float, the starting speed of the quadrotor. 
        vend, float, the ending speed of the quadrotor. 
    Outputs:
        A, numpy array, matrix of equality constraints (left side). 
        b, numpy array, array of equality constraints (right side).
        G, numpy array, matrix of inequality constraints (left side). 
        h, numpy array, array of inequality constraints (right side).

    """

    # The constraint matrices to be filled out.
    A = []
    b = []
    G = []
    h = []

    for i in range(m): # for each segment...
        # Gets the segment duration
        dt = delta_t[i]

        # Position continuity at the beginning of the segment
        A.append([0]*(k+1)*i + [1] + [0]*(k) + [0]*(k+1)*(m-i-1))
        b.append(keyframes[i])

        # Position continuity at the end of the segment
        A.append([0]*(k+1)*i + [dt**j for j in range(k+1)] + [0]*(k+1)*(m-i-1))
        b.append(keyframes[i+1])

        # Intermediate smoothness constraints
        if i < (m-1): # we don't want to include the last segment for this loop

            A.append([0]*(k+1)*i + [0] + [-j*dt**(j-1) for j in range(1, k+1)] + [0] + [j*(0)**(j-1) for j in range(1, k+1)] + [0]*(k+1)*(m-i-2))  # Velocity
            b.append(0)
            A.append([0]*(k+1)*i + [0]*2 + [-(j-1)*j*dt**(j-2) for j in range(2, k+1)] + [0]*2 + [(j-1)*j*(0)**(j-2) for j in range(2, k+1)] + [0]*(k+1)*(m-i-2))  # Acceleration
            b.append(0)
            A.append([0]*(k+1)*i + [0]*3 + [-(j-2)*(j-1)*j*dt**(j-3) for j in range(3, k+1)] + [0]*3 + [(j-2)*(j-1)*j*(0)**(j-3) for j in range(3, k+1)] + [0]*(k+1)*(m-i-2))  # Jerk
            b.append(0)
            A.append([0]*(k+1)*i + [0]*4 + [-(j-3)*(j-2)*(j-1)*j*dt**(j-4) for j in range(4, k+1)] + [0]*4 + [(j-3)*(j-2)*(j-1)*j*(0)**(j-4) for j in range(4, k+1)] + [0]*(k+1)*(m-i-2))  # Snap
            b.append(0)

        # Inequality constraints 
        G.append([0]*(k+1)*i + [0] + [j*(0.5*dt)**(j-1) for j in range(1, k+1)] + [0]*(k+1)*(m-i-1))  # Velocity constraint at midpoint
        h.append(vmax)

    A.append([0] + [j*(0)**(j-1) for j in range(1, k+1)] + [0]*(k+1)*(m-1))              # Velocity at start
    b.append(vstart)
    A.append([0]*(k+1)*(m-1) + [0] + [j*(dt)**(j-1) for j in range(1, k+1)])              # Velocity at end
    b.append(vend)
    A.append([0]*2 + [(j-1)*j*(0)**(j-2) for j in range(2, k+1)] + [0]*(k+1)*(m-1))        # Acceleration = 0 at start
    b.append(0)
    A.append([0]*(k+1)*(m-1) + [0]*2 + [(j-1)*j*(dt)**(j-2) for j in range(2, k+1)])       # Acceleration = 0 at end
    b.append(0)
    A.append([0]*3 + [(j-2)*(j-1)*j*(0)**(j-3) for j in range(3, k+1)]  + [0]*(k+1)*(m-1))        # Jerk = 0 at start
    b.append(0)
    A.append([0]*(k+1)*(m-1) + [0]*3 + [(j-2)*(j-1)*j*(dt)**(j-3) for j in range(3, k+1)])       # Jerk = 0 at end
    b.append(0)

    # Convert to numpy arrays and ensure floats to work with cvxopt. 
    A = np.array(A).astype(float)
    b = np.array(b).astype(float)
    G = np.array(G).astype(float)
    h = np.array(h).astype(float)

    return (A, b, G, h)

class MinSnap(object):
    """
    MinSnap generates a minimum snap trajectory for the quadrotor, following https://ieeexplore.ieee.org/document/5980409. 
    The trajectory is a piecewise 7th order polynomial (minimum degree necessary for snap optimality). 
    """
    def __init__(self, points, yaw_angles=None, yaw_rate_max=2*np.pi, 
                poly_degree=7, yaw_poly_degree=7,
                v_max=3, v_avg=1, v_start=[0, 0, 0], v_end=[0, 0, 0],
                verbose=True):
        """
        Waypoints and yaw angles compose the "keyframes" for optimizing over. 
        Inputs:
            points, numpy array of m 3D waypoints. 
            yaw_angles, numpy array of m yaw angles corresponding to each waypoint. 
            yaw_rate_max, the maximum yaw rate in rad/s
            v_avg, the average speed between waypoints, this is used to do the time allocation as well as impose constraints at midpoint of each segment. 
            v_start, the starting velocity vector given as an array [x_dot_start, y_dot_start, z_dot_start]
            v_end, the ending velocity vector given as an array [x_dot_end, y_dot_end, z_dot_end]
            verbose, determines whether or not the QP solver will output information. 
        """

        if poly_degree != 7 or yaw_poly_degree != 7:
            raise NotImplementedError("Oops, we haven't implemented cost functions for polynomial degree != 7 yet.")

        if yaw_angles is None:
            self.yaw = np.zeros((points.shape[0]))
        else:
            self.yaw = yaw_angles
        self.v_avg = v_avg

        cvxopt.solvers.options['show_progress'] = verbose

        # Compute the distances between each waypoint.
        seg_dist = np.linalg.norm(np.diff(points, axis=0), axis=1)
        seg_mask = np.append(True, seg_dist > 1e-1)
        self.points = points[seg_mask,:]

        self.null = False

        m = self.points.shape[0]-1  # Get the number of segments

        # Compute the derivatives of the polynomials
        self.x_dot_poly    = np.zeros((m, 3, poly_degree))
        self.x_ddot_poly   = np.zeros((m, 3, poly_degree-1))
        self.x_dddot_poly  = np.zeros((m, 3, poly_degree-2))
        self.x_ddddot_poly = np.zeros((m, 3, poly_degree-3))
        self.yaw_dot_poly = np.zeros((m, 1, yaw_poly_degree))
        self.yaw_ddot_poly = np.zeros((m, 1, yaw_poly_degree-1))
        
        # If two or more waypoints remain, solve min snap
        if self.points.shape[0] >= 2:

            ################## Time allocation
            self.delta_t = seg_dist/self.v_avg # Compute the segment durations based on the average velocity
            self.t_keyframes = np.concatenate(([0], np.cumsum(self.delta_t)))  # Construct time array which indicates when the quad should be at the i'th waypoint. 

            ################## Cost function
            # First get the cost segment for each matrix: 
            H_pos = [H_fun(self.delta_t[i], k=poly_degree) for i in range(m)]
            H_yaw = [H_fun(self.delta_t[i], k=yaw_poly_degree) for i in range(m)]

            # Now concatenate these costs using block diagonal form:
            P_pos = block_diag(*H_pos)
            P_yaw = block_diag(*H_yaw)

            # Lastly the linear term in the cost function is 0
            q_pos = np.zeros(((poly_degree+1)*m,1))
            q_yaw = np.zeros(((yaw_poly_degree+1)*m, 1))
            
            ################## Constraints for each axis
            (Ax,bx,Gx,hx) = get_1d_constraints(self.points[:,0], self.delta_t, m, k=poly_degree, vmax=v_max, vstart=v_start[0], vend=v_end[0])
            (Ay,by,Gy,hy) = get_1d_constraints(self.points[:,1], self.delta_t, m, k=poly_degree, vmax=v_max, vstart=v_start[1], vend=v_end[1])
            (Az,bz,Gz,hz) = get_1d_constraints(self.points[:,2], self.delta_t, m, k=poly_degree, vmax=v_max, vstart=v_start[2], vend=v_end[2])
            (Ayaw,byaw,Gyaw,hyaw) = get_1d_constraints(self.yaw, self.delta_t, m, k=yaw_poly_degree, vmax=yaw_rate_max)

            ################## Solve for x, y, z, and yaw

            ### Only in the fully constrained situation is there a unique minimum s.t. we can solve the system Ax = b.
            # c_opt_x = np.linalg.solve(Ax,bx)
            # c_opt_y = np.linalg.solve(Ay,by)
            # c_opt_z = np.linalg.solve(Az,bz)
            # c_opt_yaw = np.linalg.solve(Ayaw,byaw)

            ### Otherwise, in the underconstrained case or when inequality constraints are given we solve the QP. 
            c_opt_x = cvxopt_solve_qp(P_pos, q=q_pos, G=Gx, h=hx, A=Ax, b=bx)
            c_opt_y = cvxopt_solve_qp(P_pos, q=q_pos, G=Gy, h=hy, A=Ay, b=by)
            c_opt_z = cvxopt_solve_qp(P_pos, q=q_pos, G=Gz, h=hz, A=Az, b=bz)
            c_opt_yaw = cvxopt_solve_qp(P_yaw, q=q_yaw, G=Gyaw, h=hyaw, A=Ayaw, b=byaw)

            ################## Construct polynomials from c_opt
            self.x_poly = np.zeros((m, 3, (poly_degree+1)))
            self.yaw_poly = np.zeros((m, 1, (yaw_poly_degree+1)))
            for i in range(m):
                self.x_poly[i,0,:] = np.flip(c_opt_x[(poly_degree+1)*i:((poly_degree+1)*i+(poly_degree+1))])
                self.x_poly[i,1,:] = np.flip(c_opt_y[(poly_degree+1)*i:((poly_degree+1)*i+(poly_degree+1))])
                self.x_poly[i,2,:] = np.flip(c_opt_z[(poly_degree+1)*i:((poly_degree+1)*i+(poly_degree+1))])
                self.yaw_poly[i,0,:] = np.flip(c_opt_yaw[(yaw_poly_degree+1)*i:((yaw_poly_degree+1)*i+(yaw_poly_degree+1))])

            for i in range(m):
                for j in range(3):
                    self.x_dot_poly[i,j,:]    = np.polyder(self.x_poly[i,j,:], m=1)
                    self.x_ddot_poly[i,j,:]   = np.polyder(self.x_poly[i,j,:], m=2)
                    self.x_dddot_poly[i,j,:]  = np.polyder(self.x_poly[i,j,:], m=3)
                    self.x_ddddot_poly[i,j,:] = np.polyder(self.x_poly[i,j,:], m=4)
                self.yaw_dot_poly[i,0,:] = np.polyder(self.yaw_poly[i,0,:], m=1)
                self.yaw_ddot_poly[i,0,:] = np.polyder(self.yaw_poly[i,0,:], m=2)

        else:
            # Otherwise, there is only one waypoint so we just set everything = 0. 
            self.null = True
            m = 1
            self.T = np.zeros((m,))
            self.x_poly = np.zeros((m, 3, 6))
            self.x_poly[0,:,-1] = points[0,:]

    def update(self, t):
        """
        Given the present time, return the desired flat output and derivatives.

        Inputs
            t, time, s
        Outputs
            flat_output, a dict describing the present desired flat outputs with keys
                x,        position, m
                x_dot,    velocity, m/s
                x_ddot,   acceleration, m/s**2
                x_dddot,  jerk, m/s**3
                x_ddddot, snap, m/s**4
                yaw,      yaw angle, rad
                yaw_dot,  yaw rate, rad/s
        """
        x = np.zeros((3,))
        x_dot = np.zeros((3,))
        x_ddot = np.zeros((3,))
        x_dddot = np.zeros((3,))
        x_ddddot = np.zeros((3,))
        yaw    = 0
        yaw_dot = 0
        yaw_ddot = 0

        if self.null:
            # If there's only one waypoint
            x    = self.points[0,:]
            yaw = self.yaw[0]
        else:
            # Find interval index i and time within interval t.
            t = np.clip(t, self.t_keyframes[0], self.t_keyframes[-1])
            for i in range(self.t_keyframes.size-1):
                if self.t_keyframes[i] + self.delta_t[i] >= t:
                    break
            t = t - self.t_keyframes[i]

            # Evaluate polynomial.
            for j in range(3):
                x[j]        = np.polyval(       self.x_poly[i,j,:], t)
                x_dot[j]    = np.polyval(   self.x_dot_poly[i,j,:], t)
                x_ddot[j]   = np.polyval(  self.x_ddot_poly[i,j,:], t)
                x_dddot[j]  = np.polyval( self.x_dddot_poly[i,j,:], t)
                x_ddddot[j] = np.polyval(self.x_ddddot_poly[i,j,:], t)

            yaw = np.polyval(self.yaw_poly[i, 0, :], t)
            yaw_dot = np.polyval(self.yaw_dot_poly[i,0,:], t)
            yaw_ddot = np.polyval(self.yaw_ddot_poly[i,0,:], t)

        flat_output = { 'x':x, 'x_dot':x_dot, 'x_ddot':x_ddot, 'x_dddot':x_dddot, 'x_ddddot':x_ddddot,
                        'yaw':yaw, 'yaw_dot':yaw_dot, 'yaw_ddot':yaw_ddot}
        return flat_output

if __name__=="__main__":

    import matplotlib.pyplot as plt
    from matplotlib import cm

    waypoints = np.array([[0.00, 0.00, 0.00],
                          [1.00, 0.00, 0.25],
                          [1.00, 1.00, 0.50],
                          [0.00, 1.00, 1.00],
                          [0.00, 2.00, 1.25],
                          [2.00, 2.00, 1.50]
                          ])
    yaw_angles = np.array([0, np.pi/2, 0, np.pi/4, 3*np.pi/2, 0])
    v_avg = 2                                                       # Average velocity, used for time allocation
    v_start = [0, 0, 0]                                             # Start (x,y,z) velocity
    v_end = [0, 0, 0]                                               # End (x,y,z) velocity.

    traj = MinSnap(waypoints, yaw_angles, v_max=5, v_avg=v_avg, v_start=v_start, v_end=v_end, verbose=True)
    t_keyframes = traj.t_keyframes

    N = 1000
    time = np.linspace(0,1.1*t_keyframes[-1],N)
    x = np.zeros((N, 3))
    xdot = np.zeros((N,3))
    xddot = np.zeros((N,3))
    xdddot = np.zeros((N,3))
    xddddot = np.zeros((N,3))
    yaw = np.zeros((N,))
    yaw_dot = np.zeros((N,))

    for i in range(N):
        flat = traj.update(time[i])

        x[i,:] = flat['x']
        xdot[i,:] = flat['x_dot']
        xddot[i,:] = flat['x_ddot']
        xdddot[i,:] = flat['x_dddot']
        xddddot[i,:] = flat['x_ddddot']

        yaw[i] = flat['yaw']
        yaw_dot[i] = flat['yaw_dot']

    t_keyframes = traj.t_keyframes

    (fig, axes) = plt.subplots(nrows=5, ncols=1, sharex=True, num="Translational Flat Outputs")
    ax = axes[0]
    ax.plot(time, x[:,0], 'r-', label="X")
    ax.plot(time, x[:,1], 'g-', label="Y")
    ax.plot(time, x[:,2], 'b-', label="Z")
    ax.plot(t_keyframes, waypoints, 'ko')
    ax.legend()
    ax.set_ylabel("x")
    ax = axes[1]
    ax.plot(time, xdot[:,0], 'r-', label="X")
    ax.plot(time, xdot[:,1], 'g-', label="Y")
    ax.plot(time, xdot[:,2], 'b-', label="Z")
    ax.set_ylabel("xdot")
    ax = axes[2]
    ax.plot(time, xddot[:,0], 'r-', label="X")
    ax.plot(time, xddot[:,1], 'g-', label="Y")
    ax.plot(time, xddot[:,2], 'b-', label="Z")
    ax.set_ylabel("xddot")
    ax = axes[3]
    ax.plot(time, xdddot[:,0], 'r-', label="X")
    ax.plot(time, xdddot[:,1], 'g-', label="Y")
    ax.plot(time, xdddot[:,2], 'b-', label="Z")
    ax.set_ylabel("xdddot")
    ax = axes[4]
    ax.plot(time, xddddot[:,0], 'r-', label="X")
    ax.plot(time, xddddot[:,1], 'g-', label="Y")
    ax.plot(time, xddddot[:,2], 'b-', label="Z")
    ax.set_ylabel("xddddot")
    ax.set_xlabel("Time, s")

    (fig, axes) = plt.subplots(nrows=2, ncols=1, sharex=True, num="Yaw vs Time")
    ax = axes[0]
    ax.plot(time, yaw, 'k', label="yaw")
    ax.plot(t_keyframes, yaw_angles, 'ro')
    ax.set_ylabel("yaw")
    ax = axes[1]
    ax.plot(time, yaw_dot, 'k', label="yaw")
    ax.set_ylabel("yaw dot")
    ax.set_xlabel("Time, s")

    speed = np.sqrt(xdot[:,0]**2 + xdot[:,1]**2)
    fig = plt.figure(num="XY Trajectory")
    ax = fig.add_subplot(projection='3d')
    s = ax.scatter(x[:,0], x[:,1], x[:,2], c=cm.winter(speed/speed.max()), s=4, label="Flat Output")
    ax.plot(waypoints[:,0], waypoints[:,1], waypoints[:,2], 'ko', markersize=10, label="Waypoints")
    ax.grid()
    ax.legend()

    plt.show()


import numpy as np

class Polynomial(object):
    """

    """
    def __init__(self, points, v_avg=1.2):
        """
        Inputs:
            points, (N, 3) array of N waypoint coordinates in 3D
            v_avg, the average speed between segments
        """

        def get_poly(xi, xf, T):
            """
            Return fully constrained polynomial coefficients from xi to xf in
            time interval [0,T]. Low derivatives are all zero at the endpoints.
            """
            A = np.array([[      0,       0,      0,    0, 0, 1],
                          [   T**5,    T**4,   T**3, T**2, T, 1],
                          [      0,       0,      0,    0, 1, 0],
                          [ 5*T**4,  4*T**3, 3*T**2,  2*T, 1, 0],
                          [      0,       0,      0,    2, 0, 0],
                          [20*T**3, 12*T**2,    6*T,    2, 0, 0]])
            b = np.array([xi, xf, 0, 0, 0, 0])
            poly = np.linalg.solve(A,b)
            return poly

        self.v_avg = v_avg

        # Remove any sequential duplicate points; always keep the first point.
        seg_dist = np.linalg.norm(np.diff(points, axis=0), axis=1)
        seg_mask = np.append(True, seg_dist > 1e-3)
        points = points[seg_mask,:]

        # If at least two waypoints remain, calculate segment polynomials.
        if points.shape[0] >= 2:
            N = points.shape[0]-1
            self.T = seg_dist/self.v_avg # segment duration
            self.x_poly = np.zeros((N, 3, 6))
            for i in range(N):
                for j in range(3):
                    self.x_poly[i,j,:] = get_poly(points[i,j], points[i+1,j], self.T[i])
        # Otherwise, hard code constant polynomial at initial waypoint position.
        else:
            N = 1
            self.T = np.zeros((N,))
            self.x_poly = np.zeros((N, 3, 6))
            self.x_poly[0,:,-1] = points[0,:]

        # Calculate global start time of each segment.
        self.t_start = np.concatenate(([0], np.cumsum(self.T[:-1])))

        # Calculate derivative polynomials.
        self.x_dot_poly    = np.zeros((N, 3, 5))
        self.x_ddot_poly   = np.zeros((N, 3, 4))
        self.x_dddot_poly  = np.zeros((N, 3, 3))
        self.x_ddddot_poly = np.zeros((N, 3, 2))
        for i in range(N):
            for j in range(3):
                self.x_dot_poly[i,j,:]    = np.polyder(self.x_poly[i,j,:], m=1)
                self.x_ddot_poly[i,j,:]   = np.polyder(self.x_poly[i,j,:], m=2)
                self.x_dddot_poly[i,j,:]  = np.polyder(self.x_poly[i,j,:], m=3)
                self.x_ddddot_poly[i,j,:] = np.polyder(self.x_poly[i,j,:], m=4)

    def update(self, t):
        """
        Given the present time, return the desired flat output and derivatives.

        Inputs
            t, time, s
        Outputs
            flat_output, a dict describing the present desired flat outputs with keys
                x,        position, m
                x_dot,    velocity, m/s
                x_ddot,   acceleration, m/s**2
                x_dddot,  jerk, m/s**3
                x_ddddot, snap, m/s**4
                yaw,      yaw angle, rad
                yaw_dot,  yaw rate, rad/s
        """
        x        = np.zeros((3,))
        x_dot    = np.zeros((3,))
        x_ddot   = np.zeros((3,))
        x_dddot  = np.zeros((3,))
        x_ddddot = np.zeros((3,))
        yaw = 0
        yaw_dot = 0
        yaw_ddot = 0

        # Find interval index i and time within interval t.
        t = np.clip(t, self.t_start[0], self.t_start[-1]+self.T[-1])
        for i in range(self.t_start.size):
            if self.t_start[i] + self.T[i] >= t:
                break
        t = t - self.t_start[i]

        # Evaluate polynomial.
        for j in range(3):
            x[j]        = np.polyval(       self.x_poly[i,j,:], t)
            x_dot[j]    = np.polyval(   self.x_dot_poly[i,j,:], t)
            x_ddot[j]   = np.polyval(  self.x_ddot_poly[i,j,:], t)
            x_dddot[j]  = np.polyval( self.x_dddot_poly[i,j,:], t)
            x_ddddot[j] = np.polyval(self.x_ddddot_poly[i,j,:], t)

        flat_output = { 'x':x, 'x_dot':x_dot, 'x_ddot':x_ddot, 'x_dddot':x_dddot, 'x_ddddot':x_ddddot,
                        'yaw':yaw, 'yaw_dot':yaw_dot, 'yaw_ddot':yaw_ddot}
        return flat_output



import pandas as pd
import numpy as np
from scipy.spatial.transform import Rotation
import os


def unpack_sim_data(result):
    """
    Unpacks the simulated result dictionary and converts it to a Pandas DataFrame

    """
    
    headers = [ 'time',                                                                                                                                                         # Time
            'x', 'y', 'z', 'xdot', 'ydot', 'zdot', 'qx', 'qy', 'qz', 'qw', 'wx', 'wy', 'wz', 'windx', 'windy', 'windz', 'r1', 'r2', 'r3', 'r4',                             # Vehicle state                                                                                                                                               # GT body velocity
            'xdes', 'ydes', 'zdes', 'xdotdes', 'ydotdes', 'zdotdes', 'xddotdes', 'yddotdes', 'zddotdes', 'xdddotdes', 'ydddotdes', 'zdddotdes',                             # Flat outputs
            'xddddotdes', 'yddddotdes', 'zddddotdes', 'yawdes', 'yawdotdes', 
            'ax', 'ay', 'az', 'ax_gt', 'ay_gt', 'az_gt', 'gx', 'gy', 'gz',                                                                                                  # IMU measurements
            'mocap_x', 'mocap_y', 'mocap_z', 'mocap_xdot', 'mocap_ydot', 'mocap_zdot', 'mocap_qx', 'mocap_qy', 'mocap_qz', 'mocap_qw', 'mocap_wx', 'mocap_wy', 'mocap_wz',  # Mocap measurements
            'r1des', 'r2des', 'r3des', 'r4des', 'thrustdes', 'qxdes', 'qydes', 'qzdes', 'qwdes', 'mxdes', 'mydes', 'mzdes',                                                 # Controller
    ]

    # Extract data into numpy arrays
    time                = result['time'].reshape(-1,1)
    state               = result['state']
    x               = state['x']
    v               = state['v']
    q               = state['q']
    w               = state['w']
    wind            = state['wind']
    rotor_speeds    = state['rotor_speeds']
    control             = result['control']
    cmd_rotor   = control['cmd_motor_speeds']
    cmd_thrust  = control['cmd_thrust'].reshape(-1,1)
    cmd_q       = control['cmd_q']
    cmd_moment  = control['cmd_moment']

    flat                = result['flat']
    x_des       = flat['x']
    v_des       = flat['x_dot']
    a_des       = flat['x_ddot']
    j_des       = flat['x_dddot']
    s_des       = flat['x_ddddot']
    yaw_des     = flat['yaw'].reshape(-1,1)
    yawdot_des  = flat['yaw_dot'].reshape(-1,1)

    imu_measurements    = result['imu_measurements']
    a_measured = imu_measurements['accel']
    w_measured = imu_measurements['gyro']

    mocap_measurements = result['mocap_measurements']
    x_mc = mocap_measurements['x']
    v_mc = mocap_measurements['v']
    q_mc = mocap_measurements['q']
    w_mc = mocap_measurements['w']

    imu_actual          = result['imu_gt']
    a_actual = imu_actual['accel']

    state_estimate      = result['state_estimate']
    filter_state = state_estimate['filter_state']
    covariance = state_estimate['covariance']

    if filter_state.shape[1] > 0:
        # Computes the standard deviation of the filter covariance diagonal elements
        sd = 3*np.sqrt(np.diagonal(covariance, axis1=1, axis2=2))
        headers.extend(['xhat_'+str(i) for i in range(filter_state.shape[1])])
        headers.extend(['sigma_'+str(i) for i in range(filter_state.shape[1])])

        dataset = np.hstack((time,
                             x, v, q, w, wind, rotor_speeds,
                             x_des, v_des, a_des, j_des, s_des, yaw_des, yawdot_des, 
                             a_measured, a_actual, w_measured, 
                             x_mc, v_mc, q_mc, w_mc,
                             cmd_rotor, cmd_thrust, cmd_q, cmd_moment, 
                             filter_state, sd))
    else:
        sd = []

        dataset = np.hstack((time,
                            x, v, q, w, wind, rotor_speeds,
                            x_des, v_des, a_des, j_des, s_des, yaw_des, yawdot_des, 
                            a_measured, a_actual, w_measured, 
                            x_mc, v_mc, q_mc, w_mc,
                            cmd_rotor, cmd_thrust, cmd_q, cmd_moment))
    df = pd.DataFrame(data=dataset,
                        columns=headers)
    
    return df

def save_sim_data(result, filename="output.csv"):
    """
    Saves the data in result as a csv file for post processing.
    """

    if not (".csv" in filename):
        # Check whether or not filename contains .csv
        filename = filename + ".csv"

    # Get dataframe from simulation data. 
    df = unpack_sim_data(result)

    # Send it to csv file and save it in /rotorpy/data_out/ file. 
    path = os.path.join(os.path.dirname(__file__), '..', 'data_out/'+filename)
    df.to_csv(path)

    return 


import numpy as np
import sys

class CircularTraj(object):
    """
    A circle. 
    """
    def __init__(self, center=np.array([0,0,0]), radius=1, freq=0.2, yaw_bool=False, plane='XY', direction='CCW'):
        """
        This is the constructor for the Trajectory object. A fresh trajectory
        object will be constructed before each mission.

        Inputs:
            center, the center of the circle (m)
            radius, the radius of the circle (m)
            freq, the frequency with which a circle is completed (Hz)
            yaw_bool, determines if yaw motion is desired
            plane, the plane with which the circle lies on, 'XY', 'YZ', or 'XZ' 
            direcition, the direction of the circle, 'CCW' or 'CW'
        """

        # Check and assign inputs
        if plane == "XY" or plane == "YZ" or plane == "XZ":
            self.plane = plane
        else:
            print("CircularTraj Error: incorrect specification of plane. Must be 'XY', 'YZ', or 'XZ' ")
            sys.exit(1)

        if direction == "CW" or direction == "CCW":
            if direction == "CW":
                self.sign = -1
            else:
                self.sign = 1
        else:
            print("CircularTraj Error: incorrect specification of direction. Must be 'CW' or 'CCW' ")
            sys.exit(1)

        self.center = center
        self.cx, self.cy, self.cz = center[0], center[1], center[2]
        self.radius = radius
        self.freq = freq

        self.omega = 2*np.pi*self.freq
        self.yaw_bool = yaw_bool


    def update(self, t):
        """
        Given the present time, return the desired flat output and derivatives.

        Inputs
            t, time, s
        Outputs
            flat_output, a dict describing the present desired flat outputs with keys
                x,        position, m
                x_dot,    velocity, m/s
                x_ddot,   acceleration, m/s**2
                x_dddot,  jerk, m/s**3
                x_ddddot, snap, m/s**4
                yaw,      yaw angle, rad
                yaw_dot,  yaw rate, rad/s
        """

        if self.plane == "XY":
            x        = np.array([self.cx + self.radius*np.cos(self.sign*self.omega*t),
                                self.cy + self.radius*np.sin(self.sign*self.omega*t),
                                self.cz])
            x_dot    = np.array([-self.radius*self.sign*self.omega*np.sin(self.sign*self.omega*t),
                                self.radius*self.sign*self.omega*np.cos(self.sign*self.omega*t),
                                0])
            x_ddot   = np.array([-self.radius*((self.sign*self.omega)**2)*np.cos(self.sign*self.omega*t),
                                -self.radius*((self.sign*self.omega)**2)*np.sin(self.sign*self.omega*t),
                                0])
            x_dddot  = np.array([self.radius*((self.sign*self.omega)**3)*np.sin(self.sign*self.omega*t),
                                -self.radius*((self.sign*self.omega)**3)*np.cos(self.sign*self.omega*t),
                                0])
            x_ddddot = np.array([self.radius*((self.sign*self.omega)**4)*np.cos(self.sign*self.omega*t),
                                self.radius*((self.sign*self.omega)**4)*np.sin(self.sign*self.omega*t),
                                0])
        elif self.plane == "YZ":
            x        = np.array([self.cx,
                                self.cy + self.radius*np.cos(self.sign*self.omega*t),
                                self.cz + self.radius*np.sin(self.sign*self.omega*t)])
            x_dot    = np.array([0,
                                -self.radius*self.sign*self.omega*np.sin(self.sign*self.omega*t),
                                self.radius*self.sign*self.omega*np.cos(self.sign*self.omega*t)])
            x_ddot   = np.array([0,
                                -self.radius*((self.sign*self.omega)**2)*np.cos(self.sign*self.omega*t),
                                -self.radius*((self.sign*self.omega)**2)*np.sin(self.sign*self.omega*t)])
            x_dddot  = np.array([0,
                                self.radius*((self.sign*self.omega)**3)*np.sin(self.sign*self.omega*t),
                                -self.radius*((self.sign*self.omega)**3)*np.cos(self.sign*self.omega*t)])
            x_ddddot = np.array([0,
                                self.radius*((self.sign*self.omega)**4)*np.cos(self.sign*self.omega*t),
                                self.radius*((self.sign*self.omega)**4)*np.sin(self.sign*self.omega*t)])
        elif self.plane == "XZ":
            x        = np.array([self.cx + self.radius*np.cos(self.sign*self.omega*t),
                                self.cy,
                                self.cz + self.radius*np.sin(self.sign*self.omega*t)])
            x_dot    = np.array([-self.radius*self.sign*self.omega*np.sin(self.sign*self.omega*t),
                                0,
                                self.radius*self.sign*self.omega*np.cos(self.sign*self.omega*t)])
            x_ddot   = np.array([-self.radius*((self.sign*self.omega)**2)*np.cos(self.sign*self.omega*t),
                                0,
                                -self.radius*((self.sign*self.omega)**2)*np.sin(self.sign*self.omega*t)])
            x_dddot  = np.array([self.radius*((self.sign*self.omega)**3)*np.sin(self.omega*t),
                                0,
                                -self.radius*((self.sign*self.omega)**3)*np.cos(self.sign*self.omega*t)])
            x_ddddot = np.array([self.radius*((self.sign*self.omega)**4)*np.cos(self.sign*self.omega*t),
                                0,
                                self.radius*((self.sign*self.omega)**4)*np.sin(self.sign*self.omega*t)])

        if self.yaw_bool:
            yaw = np.pi/4*np.sin(np.pi*t)
            yaw_dot = np.pi*np.pi/4*np.cos(np.pi*t)
            yaw_ddot = -np.pi*np.pi*np.pi/4*np.sin(np.pi*t)

        else:
            yaw = 0
            yaw_dot = 0
            yaw_ddot = 0

        flat_output = { 'x':x, 'x_dot':x_dot, 'x_ddot':x_ddot, 'x_dddot':x_dddot, 'x_ddddot':x_ddddot,
                        'yaw':yaw, 'yaw_dot':yaw_dot, 'yaw_ddot':yaw_ddot}
        return flat_output

class ThreeDCircularTraj(object):
    """

    """
    def __init__(self, center=np.array([0,0,0]), radius=np.array([1,1,1]), freq=np.array([0.2,0.2,0.2]), yaw_bool=False):
        """
        This is the constructor for the Trajectory object. A fresh trajectory
        object will be constructed before each mission.

        Inputs:
            center, the center of the circle (m)
            radius, the radius of the circle (m)
            freq, the frequency with which a circle is completed (Hz)
        """

        self.center = center
        self.cx, self.cy, self.cz = center[0], center[1], center[2]
        self.radius = radius
        self.freq = freq

        self.omega = 2*np.pi*self.freq

        self.yaw_bool = yaw_bool

    def update(self, t):
        """
        Given the present time, return the desired flat output and derivatives.

        Inputs
            t, time, s
        Outputs
            flat_output, a dict describing the present desired flat outputs with keys
                x,        position, m
                x_dot,    velocity, m/s
                x_ddot,   acceleration, m/s**2
                x_dddot,  jerk, m/s**3
                x_ddddot, snap, m/s**4
                yaw,      yaw angle, rad
                yaw_dot,  yaw rate, rad/s
        """
        x        = np.array([self.cx + self.radius[0]*np.cos(self.omega[0]*t),
                             self.cy + self.radius[1]*np.sin(self.omega[1]*t),
                             self.cz + self.radius[2]*np.sin(self.omega[2]*t)])
        x_dot    = np.array([-self.radius[0]*self.omega[0]*np.sin(self.omega[0]*t),
                              self.radius[1]*self.omega[1]*np.cos(self.omega[1]*t),
                              self.radius[2]*self.omega[2]*np.cos(self.omega[2]*t)])
        x_ddot   = np.array([-self.radius[0]*(self.omega[0]**2)*np.cos(self.omega[0]*t),
                             -self.radius[1]*(self.omega[1]**2)*np.sin(self.omega[1]*t),
                             -self.radius[2]*(self.omega[2]**2)*np.sin(self.omega[2]*t)])
        x_dddot  = np.array([ self.radius[0]*(self.omega[0]**3)*np.sin(self.omega[0]*t),
                             -self.radius[1]*(self.omega[1]**3)*np.cos(self.omega[1]*t),
                              self.radius[2]*(self.omega[2]**3)*np.cos(self.omega[2]*t)])
        x_ddddot = np.array([self.radius[0]*(self.omega[0]**4)*np.cos(self.omega[0]*t),
                             self.radius[1]*(self.omega[1]**4)*np.sin(self.omega[1]*t),
                             self.radius[2]*(self.omega[2]**4)*np.sin(self.omega[2]*t)])

        if self.yaw_bool:
            yaw = 0.8*np.pi/2*np.sin(2.5*t)
            yaw_dot = 0.8*2.5*np.pi/2*np.cos(2.5*t)
        else:
            yaw = 0
            yaw_dot = 0

        flat_output = { 'x':x, 'x_dot':x_dot, 'x_ddot':x_ddot, 'x_dddot':x_dddot, 'x_ddddot':x_ddddot,
                        'yaw':yaw, 'yaw_dot':yaw_dot}
        return flat_output


import numpy as np
from scipy.spatial.transform import Rotation
import copy

def hat_map(s):
        """
        Given vector s in R^3, return associate skew symmetric matrix S in R^3x3
        """
        return np.array([[    0, -s[2],  s[1]],
                         [ s[2],     0, -s[0]],
                         [-s[1],  s[0],     0]])

class MotionCapture():
    """
    The external motion capture is able to provide pose and twist measurements of the vehicle. 
    Given the current ground truth state of the vehicle, it will output noisy measurements of the 
    pose and twist. Artifacts can be introduced 
    """
    def __init__(self, sampling_rate, 
                 mocap_params={'pos_noise_density': 0.0005*np.ones((3,)),  # noise density for position 
                                                    'vel_noise_density': 0.005*np.ones((3,)),          # noise density for velocity
                                                    'att_noise_density': 0.0005*np.ones((3,)),          # noise density for attitude 
                                                    'rate_noise_density': 0.0005*np.ones((3,)),         # noise density for body rates
                                                    'vel_artifact_max': 5,                              # maximum magnitude of the artifact in velocity (m/s)
                                                    'vel_artifact_prob': 0.001,                         # probability that an artifact will occur for a given velocity measurement
                                                    'rate_artifact_max': 1,                             # maximum magnitude of the artifact in body rates (rad/s)
                                                    'rate_artifact_prob': 0.0002                        # probability that an artifact will occur for a given rate measurement
                                                    }, 
                with_artifacts=False):
        """
        Parameters: 
            sampling_rate, Hz, the rate at which this sensor is being sampled. Used for computing the noise.
            mocap_params, a dict with keys
                pos_noise_density, position noise density, [m/sqrt(Hz)] 
                vel_noise_density, velocity noise density, [m/s / sqrt(Hz)]
                att_noise_density, attitude noise density, [rad / sqrt(Hz)]
                rate_noise_density, attitude rate noise density, [rad/s /sqrt(Hz)]
                vel_artifact_prob, probability that a spike will occur for a given velocity measurement
                vel_artifact_max, the maximum magnitude of the artifact spike. [m/s]
                rate_artifact_prob, probability that a spike will occur for a given body rate measurement
                rate_artifact_max, the maximum magnitude of hte artifact spike. [rad/s]
        """

        self.rate_scale = np.sqrt(sampling_rate/2)

        # Noise densities
        self.pos_density = mocap_params['pos_noise_density']
        self.vel_density = mocap_params['vel_noise_density']
        self.att_density = mocap_params['att_noise_density']
        self.rate_density = mocap_params['rate_noise_density']

        # Artifacts
        self.vel_artifact_prob = mocap_params['vel_artifact_prob']
        self.vel_artifact_max = mocap_params['vel_artifact_max']
        self.rate_artifact_prob = mocap_params['rate_artifact_prob']
        self.rate_artifact_max = mocap_params['rate_artifact_max']

        self.initialized = True
        self.with_artifacts = with_artifacts

    def measurement(self, state, with_noise=False, with_artifacts=False):
        """
        Computes and returns the sensor measurement at a time step. 
        Inputs:
            state, a dict describing the state with keys
                    x, position, m, shape=(3,)
                    v, linear velocity, m/s, shape=(3,)
                    q, quaternion [i,j,k,w], shape=(4,)
                    w, angular velocity (in LOCAL frame!), rad/s, shape=(3,)
            with_noise, a boolean to indicate if noise is added
            with_artifacts, a boolean to indicate if artifacts are added. 
                    Artifacts are added to the velocity and angular rates, and are due 
                    to the numerical differentiation scheme used by motion capture systems. 
                    They will appear as random spikes in the data. 
        Outputs:
            observation, a dictionary with keys
                    x_m, noisy position measurement, m, shape=(3,)
                    v_m, noisy linear velocity, m/s, shape=(3,)
                    q_m, noisy quaternion [i,j,k,w], shape=(4,)
                    w_m, noisy angular velocity (in LOCAL frame!), rad/s, shape=(3,)
        """
        x_measured = copy.deepcopy(state['x']).astype(float)
        v_measured = copy.deepcopy(state['v']).astype(float)
        q_measured = Rotation.from_quat(copy.deepcopy(state['q']))
        w_measured = copy.deepcopy(state['w']).astype(float)

        if with_noise:
            # Add noise to the measurements based on the provided measurement noise.
            x_measured += self.rate_scale * np.random.normal(scale=np.abs(self.pos_density))
            v_measured += self.rate_scale * np.random.normal(scale=np.abs(self.vel_density))
            w_measured += self.rate_scale * np.random.normal(scale=np.abs(self.rate_density))

            # Noise has to be treated differently with quaternions... 
            # Following https://www.iri.upc.edu/people/jsola/JoanSola/objectes/notes/kinematics.pdf  pg 43
            # First, let's produce a perturbation vector in R3
            delta_phi = self.rate_scale*np.random.normal(scale=np.abs(self.att_density))

            # Now convert that to a rotation matrix
            delta_rotation = Rotation.from_matrix(np.eye(3) + hat_map(delta_phi))

            # Now apply that rotation to the quaternion
            q_measured = (q_measured * delta_rotation).as_quat()
        else:
            q_measured = q_measured.as_quat()

        if with_artifacts:
            # If including artifacts, first roll the dice on whether or not a spike should occur for each measurement: 
            vel_spike_bool = np.random.choice([0,1], p=[1-self.vel_artifact_prob, self.vel_artifact_prob])
            rate_spike_bool = np.random.choice([0,1], p=[1-self.rate_artifact_prob, self.rate_artifact_prob])

            # Choose the axis that the spike will occur on
            vel_axis = np.random.choice([0,1,2])
            rate_axis = np.random.choice([0,1,2])

            # Choose the sign of the spike
            vel_sign = np.random.choice([-1,1])
            rate_sign = np.random.choice([-1,1])

            if vel_spike_bool:
                v_measured[vel_axis] += vel_sign*np.random.uniform(low=0, high=self.vel_artifact_max)
            if rate_spike_bool:
                w_measured[rate_axis] += rate_sign*np.random.uniform(low=0, high=self.rate_artifact_max)

        return {'x': x_measured, 'q': q_measured, 'v': v_measured, 'w': w_measured}

if __name__=="__main__":

    import matplotlib.pyplot as plt

    def merge_dicts(dicts_in):
        """
        Concatenates contents of a list of N state dicts into a single dict by
        prepending a new dimension of size N. This is more convenient for plotting
        and analysis. Requires dicts to have consistent keys and have values that
        are numpy arrays.
        """
        dict_out = {}
        for k in dicts_in[0].keys():
            dict_out[k] = []
            for d in dicts_in:
                dict_out[k].append(d[k])
            dict_out[k] = np.array(dict_out[k])
        return dict_out

    sim_rate = 1/500
    mocap_params = {'pos_noise_density': 0.0005*np.ones((3,)),  # noise density for position 
                'vel_noise_density': 0.005*np.ones((3,)),          # noise density for velocity
                'att_noise_density': 0.0005*np.ones((3,)),          # noise density for attitude 
                'rate_noise_density': 0.0005*np.ones((3,)),         # noise density for body rates
                'vel_artifact_max': 5,                              # maximum magnitude of the artifact in velocity (m/s)
                'vel_artifact_prob': 0.001,                         # probability that an artifact will occur for a given velocity measurement
                'rate_artifact_max': 1,                             # maximum magnitude of the artifact in body rates (rad/s)
                'rate_artifact_prob': 0.0002                        # probability that an artifact will occur for a given rate measurement
                }
    sensor = MotionCapture(sampling_rate=sim_rate, mocap_params=mocap_params, with_artifacts=True)

    measurements = []

    state = {'x': np.zeros((3,)), 'v': np.zeros((3,)), 'q': np.array([0,0,0,1]), 'w': np.zeros((3,))}
    for i in range(1000):
        state = {'x': np.array([np.sin(2*np.pi*i/1000), np.sin(2*np.pi*i/1000 - np.pi/2), np.sin(2*np.pi*i/1000 - np.pi/5)]), 
                 'v': np.array([np.sin(2*np.pi*i/1000), np.sin(2*np.pi*i/1000 - np.pi/2), np.sin(2*np.pi*i/1000 - np.pi/5)]), 
                 'q': np.array([0,0,0,1]), 
                 'w': np.array([np.sin(2*np.pi*i/1000), np.sin(2*np.pi*i/1000 - np.pi/2), np.sin(2*np.pi*i/1000 - np.pi/5)])}
        current = sensor.measurement(state, with_noise=True, with_artifacts=True)
        measurements.append(current)

    measurements = merge_dicts(measurements)

    x_m = measurements['x']
    v_m = measurements['v']
    q_m = measurements['q']
    w_m = measurements['w']

    q_norm = np.linalg.norm(q_m, axis=1)
    
    (fig, axes) = plt.subplots(nrows=4, ncols=1, sharex=True, num="Measurements")
    fig.set_figwidth(9)
    fig.set_figheight(9)
    axe = axes[0]
    axe.plot(x_m[:,0], 'r', markersize=2)
    axe.plot(x_m[:,1], 'g', markersize=2)
    axe.plot(x_m[:,2], 'b', markersize=2)
    axe.set_ylim(bottom=-1.5, top=1.5)
    axe = axes[1]
    axe.plot(v_m[:,0], 'r', markersize=2)
    axe.plot(v_m[:,1], 'g', markersize=2)
    axe.plot(v_m[:,2], 'b', markersize=2)
    axe.set_ylim(bottom=-1.5, top=1.5)
    axe = axes[2]
    axe.plot(q_m[:,0], 'r', markersize=2)
    axe.plot(q_m[:,1], 'g', markersize=2)
    axe.plot(q_m[:,2], 'b', markersize=2)
    axe.plot(q_m[:,3], 'm', markersize=2)
    axe.plot(q_norm, 'k', markersize=2)
    axe = axes[3]
    axe.plot(w_m[:,0], 'r', markersize=2)
    axe.plot(w_m[:,1], 'g', markersize=2)
    axe.plot(w_m[:,2], 'b', markersize=2)
    axe.set_ylim(bottom=-1.5, top=1.5)

    plt.show()


print('The below information will help debug installation problems.\n')

import sys
print('Python Version:')
print(sys.version)
print()

import platform
print('Operating system:')
print(platform.platform())
print()

import matplotlib
print('matplotlib:', matplotlib.__version__)

import numpy
print('numpy:     ', numpy.__version__)

import pip
print('pip:       ', pip.__version__)

import scipy
print('scipy:     ', scipy.__version__)



"""
Wind profiles should be implemented here to be added as an argument in the 
simulate method found in simulate.py

The main structure of each object is the only thing that must remain consistent.
When creating a new wind profile, the object must have the following:

- An __init__ method. The arguments vary depending on the wind profile. 
- An update method. The arguments *must* include time and position. 

"""
import numpy as np

class WindTemplate(object):
    """
    Winds are implemented with two required methods: __init__() and update(). 
    With __init__() you can specify any parameters or constants used to specify the wind. 
    The update() method is called at each time step of the simulator. The position and time 
    of the vehicle are provided. 
    """

    def __init__(self):
        """
        Inputs: 
            Nothing
        """

    def update(self, t, position):
        """
        Given the present time and position of the vehicle in the world frame, return the
        current wind speed on all three axes. 
        
        Be careful whether or not the wind is specified in the body or world coordinates!. 
        """
        return np.array([0,0,0])


"""
Physical parameters for a quadrotor. Values parameterize the inertia, motor dynamics, 
rotor aerodynamics, parasitic drag, and rotor placement. 
Additional sources:
    https://bitcraze.io/2015/02/measuring-propeller-rpm-part-3
    https://wiki.bitcraze.io/misc:investigations:thrust
    https://commons.erau.edu/cgi/viewcontent.cgi?article=2057&context=publication
Notes:
    k_thrust is inferred from 14.5 g thrust at 2500 rad/s
    k_drag is mostly made up
"""
import numpy as np

d = 0.043  # Arm length

quad_params = {

    # Inertial properties
    'mass': 0.03,       # kg
    'Ixx':  1.43e-5,    # kg*m^2
    'Iyy':  1.43e-5,    # kg*m^2
    'Izz':  2.89e-5,    # kg*m^2
    'Ixy':  0.0,        # kg*m^2
    'Iyz':  0.0,        # kg*m^2 
    'Ixz':  0.0,        # kg*m^2

    # Geometric properties, all vectors are relative to the center of mass.
    'num_rotors': 4,                        # for looping over each actuator
    'rotor_pos': {  
                    'r1': d*np.array([ 0.70710678118, 0.70710678118, 0]),    # Location of Rotor 1, meters
                    'r2': d*np.array([ 0.70710678118,-0.70710678118, 0]),    # Location of Rotor 2, meters
                    'r3': d*np.array([-0.70710678118,-0.70710678118, 0]),    # Location of Rotor 3, meters
                    'r4': d*np.array([-0.70710678118, 0.70710678118, 0]),    # Location of Rotor 4, meters
                 },

    'rotor_directions': np.array([1,-1,1,-1]),  # This dictates the direction of the torque for each motor. 

    'rI': np.array([0,0,0]), # location of the IMU sensor, meters

    # Frame aerodynamic properties
    'c_Dx': 0.5e-2,  # parasitic drag in body x axis, N/(m/s)**2
    'c_Dy': 0.5e-2,  # parasitic drag in body y axis, N/(m/s)**2
    'c_Dz': 1e-2,  # parasitic drag in body z axis, N/(m/s)**2

    # Rotor properties
    # See "System Identification of the Crazyflie 2.0 Nano Quadrocopter", Forster 2015.
    'k_eta': 2.3e-08,           # thrust coefficient N/(rad/s)**2
    'k_m':   7.8e-10,           # yaw moment coefficient Nm/(rad/s)**2
    'k_d':   10.2506e-07,       # rotor drag coefficient N/(rad*m/s**2) = kg/rad
    'k_z':   7.553e-07,         # induced inflow coefficient N/(rad*m/s**2) = kg/rad
    'k_flap': 0.0,              # Flapping moment coefficient Nm/(rad*m/s**2) = kg*m/rad

    # Motor properties
    'tau_m': 0.005,           # motor response time, seconds
    'rotor_speed_min': 0,       # rad/s
    'rotor_speed_max': 2500,    # rad/s
    'motor_noise_std': 100,     # rad/s

}


"""
Imports
"""
import numpy as np

class ConstantSpeed(object):
    """
    
    """
    def __init__(self, init_pos, dist=1, speed=1, axis=0, repeat=False):
        """
        Constant speed will command a step response in speed on a specified axis. The following inputs
        try to ensure that the vehicle is not commanded to go beyond the boundaries of the environment. 

        INPUTS
            init_pos := the inital position for the trajectory to begin, should be the current quadrotor's position
            dist := the length of the trajectory in meters. 
            speed := the speed of the trajectory in meters. 
            axis := the axis to travel (0 -> X, 1 -> Y, 2 -> Z)
            repeat := determines if the trajectory should repeat, where the direction is switched from its current state.
        """

        self.pt1 = init_pos
        self.dist = dist    # m
        self.speed = speed  # m/s

        # Based on the length of the trajectory (distance) and the desired speed, compute how long the desired speed should be commaned. 
        self.t_width = self.dist/self.speed  # seconds
        
        # Create a vector where the "axis'th" element is 1 and the other elements are 0. This will multiplied by the flat outputs 
        # so that the only "active" axis is the one specified by "axis"
        self.active_axis = np.zeros((3,))
        self.active_axis[axis] = 1

        self.direction = 1

        self.repeat = repeat
        self.reverse_threshold = 0.01

        self.pt2 = self.pt1 + self.speed*self.t_width*self.active_axis

    def update(self, t):
        """
        Given the present time, return the desired flat output and derivatives.

        Inputs
            t, time, s
        Outputs
            flat_output, a dict describing the present desired flat outputs with keys
                x,        position, m
                x_dot,    velocity, m/s
                x_ddot,   acceleration, m/s**2
                x_dddot,  jerk, m/s**3
                x_ddddot, snap, m/s**4
                yaw,      yaw angle, rad
                yaw_dot,  yaw rate, rad/s
        """

        v = (self.speed)*self.active_axis*self.direction
        if t <= self.t_width:
            x       = self.pt1 + v*t
            x_dot   = v
        else:
            x       = self.pt2
            x_dot   = np.zeros((3,))

        x_ddot = np.zeros((3,))
        x_dddot = np.zeros((3,))
        x_ddddot = np.zeros((3,))

        yaw    = 0
        yaw_dot = 0
        yaw_ddot = 0

        flat_output = { 'x':x, 'x_dot':x_dot, 'x_ddot':x_ddot, 'x_dddot':x_dddot, 'x_ddddot':x_ddddot,
                        'yaw':yaw, 'yaw_dot':yaw_dot, 'yaw_ddot':yaw_ddot}
        return flat_output

if __name__=="__main__":

    import matplotlib.pyplot as plt

    traj = ConstantSpeed(init_pos=np.array([0,0,0]))


    t = np.linspace(0,10,500)
    x = np.zeros((t.shape[0], 3))
    
    for i in range(t.shape[0]):
        flat = traj.update(t[i])
        x[i,:] = flat['x']

    (fig, axes) = plt.subplots(nrows=3, ncols=1, num="Constant Speed Trajectory")
    ax = axes[0]
    ax.plot(t, x[:,0], 'r', linewidth=1.5)
    ax.set_ylabel("X, m")
    ax = axes[1]
    ax.plot(t, x[:,1], 'g', linewidth=1.5)
    ax.set_ylabel("Y, m")
    ax = axes[2]
    ax.plot(t, x[:,2], 'b', linewidth=1.5)
    ax.set_ylabel("Z, m")
    ax.set_xlabel("Time, s")
    
    plt.show()
    


"""
TODO: Set up figure for appropriate target video size (eg. 720p).
TODO: Decide which additional user options should be available.
"""

from datetime import datetime
from pathlib import Path

import numpy as np
from matplotlib.animation import FuncAnimation
import matplotlib.pyplot as plt
from scipy.spatial.transform import Rotation

from rotorpy.utils.shapes import Quadrotor

import os

class ClosingFuncAnimation(FuncAnimation):
    def __init__(self, fig, func, *args, **kwargs):
        self._close_on_finish = kwargs.pop('close_on_finish')
        FuncAnimation.__init__(self, fig, func, *args, **kwargs)

    # def _stop(self, *args):
    #     super()._stop(self, *args)
    #     if self._close_on_finish:
    #         plt.close(self._fig)

    def _step(self, *args):
        still_going = FuncAnimation._step(self, *args)
        if self._close_on_finish and not still_going:
            plt.close(self._fig)

def _decimate_index(time, sample_time):
    """
    Given sorted lists of source times and sample times, return indices of
    source time closest to each sample time.
    """
    index = np.arange(time.size)
    sample_index = np.round(np.interp(sample_time, time, index)).astype(int)
    return sample_index

def animate(time, position, rotation, wind, animate_wind, world, filename=None, blit=False, show_axes=True, close_on_finish=False):
    """
    Animate a completed simulation result based on the time, position, and
    rotation history. The animation may be viewed live or saved to a .mp4 video
    (slower, requires additional libraries).

    For a live view, it is absolutely critical to retain a reference to the
    returned object in order to prevent garbage collection before the animation
    has completed displaying.

    Below, M corresponds to the number of drones you're animating. If M is None, i.e. the arrays are (N,3) and (N,3,3), then it is assumed that there is only one drone.
    Otherwise, we iterate over the M drones and animate them on the same axes.

    N is the number of time steps in the simulation.

    Parameters
        time, (N,) with uniform intervals
        position, (N,M,3)
        rotation, (N,M,3,3)
        wind, (N,M,3) world wind velocity
        animate_wind, if True animate wind vector
        world, a World object
        filename, for saved video, or live view if None
        blit, if True use blit for faster animation, default is False
        show_axes, if True plot axes, default is True
        close_on_finish, if True close figure at end of live animation or save, default is False
    """

    # Check if there is only one drone.
    if len(position.shape) == 2:
        position = np.expand_dims(position, axis=1)
        rotation = np.expand_dims(rotation, axis=1)
        wind = np.expand_dims(wind, axis=1)
    M = position.shape[1]

    # Temporal style.
    rtf = 1.0 # real time factor > 1.0 is faster than real time playback
    render_fps = 30

    # Normalize the wind by the max of the wind magnitude on each axis, so that the maximum length of the arrow is decided by the scale factor
    wind_mag = np.max(np.linalg.norm(wind, axis=-1), axis=1)             # Get the wind magnitude time series
    max_wind = np.max(wind_mag)                         # Find the maximum wind magnitude in the time series

    if max_wind != 0:
        wind_arrow_scale_factor = 1                         # Scale factor for the wind arrow
        wind = wind_arrow_scale_factor*wind / max_wind

    # Decimate data to render interval; always include t=0.
    if time[-1] != 0:
        sample_time = np.arange(0, time[-1], 1/render_fps * rtf)
    else:
        sample_time = np.zeros((1,))
    index = _decimate_index(time, sample_time)
    time = time[index]
    position = position[index,:]
    rotation = rotation[index,:]
    wind = wind[index,:]

    # Set up axes.
    if filename is not None:
        if isinstance(filename, Path):
            fig = plt.figure(filename.name)
        else:
            fig = plt.figure(filename)
    else:
        fig = plt.figure('Animation')
    fig.clear()
    ax = fig.add_subplot(projection='3d')
    if not show_axes:
        ax.set_axis_off()

    quads = [Quadrotor(ax, wind=animate_wind, wind_scale_factor=1) for _ in range(M)]

    world_artists = world.draw(ax)

    title_artist = ax.set_title('t = {}'.format(time[0]))

    def init():
        ax.draw(fig.canvas.get_renderer())
        # return world_artists + list(cquad.artists) + [title_artist]
        return world_artists + [title_artist] + [q.artists for q in quads]

    def update(frame):
        title_artist.set_text('t = {:.2f}'.format(time[frame]))
        for i, quad in enumerate(quads):
            quad.transform(position=position[frame,i,:], rotation=rotation[frame,i,:,:], wind=wind[frame,i,:])
        # [a.do_3d_projection(fig.canvas.get_renderer()) for a in quad.artists]   # No longer necessary in newer matplotlib?
        # return world_artists + list(quad.artists) + [title_artist]
        return world_artists + [title_artist] + [q.artists for q in quads]

    ani = ClosingFuncAnimation(fig=fig,
                        func=update,
                        frames=time.size,
                        init_func=init,
                        interval=1000.0/render_fps,
                        repeat=False,
                        blit=blit,
                        close_on_finish=close_on_finish)

    if filename is not None:
        print('Saving Animation')
        if not ".mp4" in filename:
            filename = filename + ".mp4"
        path = os.path.join(os.path.dirname(__file__),'..','data_out',filename)
        ani.save(path,
                 writer='ffmpeg',
                 fps=render_fps,
                 dpi=100)
        if close_on_finish:
            plt.close(fig)
            ani = None

    return ani

if __name__ == "__main__":

    from rotorpy.vehicles.crazyflie_params import quad_params  # Import quad params for the quadrotor environment.
    from rotorpy.learning.quadrotor_environments import QuadrotorEnv
    from rotorpy.controllers.quadrotor_control import SE3Control
    from rotorpy.trajectories.circular_traj import CircularTraj

    import gymnasium as gym

    output_video_dir = os.path.join(os.path.dirname(__file__), "..", "data_out", "test_animation.mp4")

    # Create M SE3 drones.
    M = 3
    baseline_controller = SE3Control(quad_params) 

    def make_env():
        return gym.make("Quadrotor-v0", 
                    control_mode ='cmd_motor_speeds', 
                    quad_params = quad_params,
                    max_time = 5,
                    world = None,
                    sim_rate = 100,
                    render_mode='3D',
                    render_fps = 60,
                    color='b')

    envs = [make_env() for _ in range(M)]

    # For each environment command it to do a circle with random radius and location. 
    trajs = []
    for env in envs:
        center = np.random.uniform(low=-2, high=2, size=3)
        radius = np.random.uniform(low=0.5, high=1.5)
        freq = np.random.uniform(low=0.1, high=0.3)
        plane = np.random.choice(['XY', 'YZ', 'XZ'])
        traj = CircularTraj(center=center, radius=radius, freq=freq, plane=plane, direction=np.random.choice(['CW', 'CCW']))
        trajs.append(traj)

    # Collect observations for each environment.
    observations = [env.reset(initial_state='random')[0] for env in envs]

    # This is a list of env termination conditions so that the loop only ends when the final env is terminated. 
    terminated = [False]*len(observations)

    # Arrays for animating. 
    T = [0]
    position = [[obs[0:3] for obs in observations]]
    quat = [[obs[6:10] for obs in observations]]

    while not all(terminated):
        for (i, env) in enumerate(envs):  # For each environment...
            # Unpack the observation from the environment
            state = {'x': observations[i][0:3], 'v': observations[i][3:6], 'q': observations[i][6:10], 'w': observations[i][10:13]}

            # Command the quad to do circles.
            flat = trajs[i].update(env.unwrapped.t)
            control_dict = baseline_controller.update(0, state, flat)

            # Extract the commanded motor speeds.
            cmd_motor_speeds = control_dict['cmd_motor_speeds']

            # The environment expects the control inputs to all be within the range [-1,1]
            action = np.interp(cmd_motor_speeds, [env.unwrapped.rotor_speed_min, env.unwrapped.rotor_speed_max], [-1,1])

            # For the last environment, append the current timestep.
            if i == 0: 
                T.append(env.unwrapped.t)

            # Step the environment forward. 
            observations[i], reward, terminated[i], truncated, info = env.step(action)

        # Append arrays for plotting.
        position.append([obs[0:3] for obs in observations])
        quat.append([obs[6:10] for obs in observations])

    # Convert to numpy arrays.
    T = np.array(T)
    position = np.array(position)
    quat = np.array(quat)

    # Convert the quaternion to rotation matrix.
    rotation = np.array([Rotation.from_quat(quat[i]).as_matrix() for i in range(T.size)])

    # Animate the results.
    ani = animate(T, position, rotation, wind=np.zeros((T.size,M,3)), animate_wind=False, world=envs[0].world, filename=output_video_dir, blit=False, show_axes=True, close_on_finish=True)


import numpy as np
from scipy.spatial.transform import Rotation

class SE3Control(object):
    """
    Quadrotor trajectory tracking controller based on https://ieeexplore.ieee.org/document/5717652 

    """
    def __init__(self, quad_params):
        """
        Parameters:
            quad_params, dict with keys specified in rotorpy/vehicles
        """

        # Quadrotor physical parameters.
        # Inertial parameters
        self.mass            = quad_params['mass'] # kg
        self.Ixx             = quad_params['Ixx']  # kg*m^2
        self.Iyy             = quad_params['Iyy']  # kg*m^2
        self.Izz             = quad_params['Izz']  # kg*m^2
        self.Ixy             = quad_params['Ixy']  # kg*m^2
        self.Ixz             = quad_params['Ixz']  # kg*m^2
        self.Iyz             = quad_params['Iyz']  # kg*m^2

        # Frame parameters
        self.c_Dx            = quad_params['c_Dx']  # drag coeff, N/(m/s)**2
        self.c_Dy            = quad_params['c_Dy']  # drag coeff, N/(m/s)**2
        self.c_Dz            = quad_params['c_Dz']  # drag coeff, N/(m/s)**2

        self.num_rotors      = quad_params['num_rotors']
        self.rotor_pos       = quad_params['rotor_pos']
        self.rotor_dir       = quad_params['rotor_directions']

        # Rotor parameters    
        self.rotor_speed_min = quad_params['rotor_speed_min'] # rad/s
        self.rotor_speed_max = quad_params['rotor_speed_max'] # rad/s

        self.k_eta           = quad_params['k_eta']     # thrust coeff, N/(rad/s)**2
        self.k_m             = quad_params['k_m']       # yaw moment coeff, Nm/(rad/s)**2
        self.k_d             = quad_params['k_d']       # rotor drag coeff, N/(m/s)
        self.k_z             = quad_params['k_z']       # induced inflow coeff N/(m/s)
        self.k_flap          = quad_params['k_flap']    # Flapping moment coefficient Nm/(m/s)

        # Motor parameters
        self.tau_m           = quad_params['tau_m']     # motor reponse time, seconds

        # You may define any additional constants you like including control gains.
        self.inertia = np.array([[self.Ixx, self.Ixy, self.Ixz],
                                 [self.Ixy, self.Iyy, self.Iyz],
                                 [self.Ixz, self.Iyz, self.Izz]]) # kg*m^2
        self.g = 9.81 # m/s^2

        # Gains  
        self.kp_pos = np.array([6.5,6.5,15])
        self.kd_pos = np.array([4.0, 4.0, 9])
        self.kp_att = 544
        self.kd_att = 46.64
        self.kp_vel = 0.1*self.kp_pos   # P gain for velocity controller (only used when the control abstraction is cmd_vel)

        # Linear map from individual rotor forces to scalar thrust and vector
        # moment applied to the vehicle.
        k = self.k_m/self.k_eta  # Ratio of torque to thrust coefficient. 

        # Below is an automated generation of the control allocator matrix. It assumes that all thrust vectors are aligned
        # with the z axis and that the "sign" of each rotor yaw moment alternates starting with positive for r1. 'TM' = "thrust and moments"
        self.f_to_TM = np.vstack((np.ones((1,self.num_rotors)),
                                  np.hstack([np.cross(self.rotor_pos[key],np.array([0,0,1])).reshape(-1,1)[0:2] for key in self.rotor_pos]), 
                                 (k * self.rotor_dir).reshape(1,-1)))
        self.TM_to_f = np.linalg.inv(self.f_to_TM)
    
    def update(self, t, state, flat_output):
        """
        This function receives the current time, true state, and desired flat
        outputs. It returns the command inputs.

        Inputs:
            t, present time in seconds
            state, a dict describing the present state with keys
                x, position, m
                v, linear velocity, m/s
                q, quaternion [i,j,k,w]
                w, angular velocity, rad/s
            flat_output, a dict describing the present desired flat outputs with keys
                x,        position, m
                x_dot,    velocity, m/s
                x_ddot,   acceleration, m/s**2
                x_dddot,  jerk, m/s**3
                x_ddddot, snap, m/s**4
                yaw,      yaw angle, rad
                yaw_dot,  yaw rate, rad/s

        Outputs:
            control_input, a dict describing the present computed control inputs with keys
                cmd_motor_speeds, rad/s
                cmd_motor_thrusts, N
                cmd_thrust, N 
                cmd_moment, N*m
                cmd_q, quaternion [i,j,k,w]
                cmd_w, angular rates in the body frame, rad/s
                cmd_v, velocity in the world frame, m/s
                cmd_acc, mass normalized thrust vector in the world frame, m/s/s.

                Not all keys are used, it depends on the control_abstraction selected when initializing the Multirotor object. 
        """
        cmd_motor_speeds = np.zeros((4,))
        cmd_thrust = 0
        cmd_moment = np.zeros((3,))
        cmd_q = np.zeros((4,))

        def normalize(x):
            """Return normalized vector."""
            return x / np.linalg.norm(x)

        def vee_map(S):
            """Return vector corresponding to given skew symmetric matrix."""
            return np.array([-S[1,2], S[0,2], -S[0,1]])

        # Get the desired force vector.
        pos_err  = state['x'] - flat_output['x']
        dpos_err = state['v'] - flat_output['x_dot']
        F_des = self.mass * (- self.kp_pos*pos_err
                             - self.kd_pos*dpos_err
                             + flat_output['x_ddot']
                             + np.array([0, 0, self.g]))

        # Desired thrust is force projects onto b3 axis.
        R = Rotation.from_quat(state['q']).as_matrix()
        b3 = R @ np.array([0, 0, 1])
        u1 = np.dot(F_des, b3)

        # Desired orientation to obtain force vector.
        b3_des = normalize(F_des)
        yaw_des = flat_output['yaw']
        c1_des = np.array([np.cos(yaw_des), np.sin(yaw_des), 0])
        b2_des = normalize(np.cross(b3_des, c1_des))
        b1_des = np.cross(b2_des, b3_des)
        R_des = np.stack([b1_des, b2_des, b3_des]).T

        # Orientation error.
        S_err = 0.5 * (R_des.T @ R - R.T @ R_des)
        att_err = vee_map(S_err)

        # Angular velocity error (this is oversimplified).
        w_des = np.array([0, 0, flat_output['yaw_dot']])
        w_err = state['w'] - w_des

        # Desired torque, in units N-m.
        u2 = self.inertia @ (-self.kp_att*att_err - self.kd_att*w_err) + np.cross(state['w'], self.inertia@state['w'])  # Includes compensation for wxJw component

        # Compute command body rates by doing PD on the attitude error. 
        cmd_w = -self.kp_att*att_err - self.kd_att*w_err

        # Compute motor speeds. Avoid taking square root of negative numbers.
        TM = np.array([u1, u2[0], u2[1], u2[2]])
        cmd_rotor_thrusts = self.TM_to_f @ TM
        cmd_motor_speeds = cmd_rotor_thrusts / self.k_eta
        cmd_motor_speeds = np.sign(cmd_motor_speeds) * np.sqrt(np.abs(cmd_motor_speeds))

        # Assign controller commands.
        cmd_thrust = u1                                             # Commanded thrust, in units N.
        cmd_moment = u2                                             # Commanded moment, in units N-m.
        cmd_q = Rotation.from_matrix(R_des).as_quat()               # Commanded attitude as a quaternion.
        cmd_v = -self.kp_vel*pos_err + flat_output['x_dot']         # Commanded velocity in world frame (if using cmd_vel control abstraction), in units m/s
        cmd_acc = F_des/self.mass                                   # Commanded acceleration in world frame (if using cmd_acc control abstraction)

        control_input = {'cmd_motor_speeds':cmd_motor_speeds,
                         'cmd_motor_thrusts':cmd_rotor_thrusts,
                         'cmd_thrust':cmd_thrust,
                         'cmd_moment':cmd_moment,
                         'cmd_q':cmd_q,
                         'cmd_w':cmd_w,
                         'cmd_v':cmd_v,
                         'cmd_acc': cmd_acc}
        
        return control_input


import numpy as np
from scipy.spatial.transform import Rotation
import copy

from filterpy.kalman import UnscentedKalmanFilter
from filterpy.kalman import MerweScaledSigmaPoints

"""
The Wind UKF uses the same model as the EKF found in wind_ekf.py, but instead applies the Unscented Kalman Filter. The benefit
of this approach is the accuracy of the UKF is 3rd order (compared to EKF's 1st order), and Jacobians do not need to be computed. 
"""
class WindUKF:
    """
    WindUKF
        Given approximate dynamics near level flight, the wind EKF will produce an estimate of the local wind vector acting on the body. 
        It requires knowledge of the effective drag coefficient on each axis, which would be determined either from real flight or computed in simulation, and the mass of the vehicle. 
        The inputs to the filter are the mass normalized collective thrust and the body rates on each axis. 
        Measurements of body velocity, Euler angles, and acceleration are provided to the vehicle. 
        The filter estimates the Euler angles, body velocities, and wind velocities.

        State space: 
            xhat = [psi, theta, phi, u, v, w, windx, windy, windz]
        Measurement space: 
            u = [T/m, p, q, r]
    """

    def __init__(self, quad_params,
                       Q=np.diag(np.concatenate([0.05*np.ones(3),0.07*np.ones(3),0.01*np.ones(3)])),
                       R=np.diag(np.concatenate([0.0005*np.ones(3),0.0010*np.ones(3),np.sqrt(100/2)*(0.38**2)*np.ones(3)])), 
                       xhat0=np.array([0,0,0, 0.1,0.05,0.02, 1.5,1.5,1.5]), 
                       P0=1*np.eye(9),
                       dt=1/100,
                       alpha=0.1,
                       beta=2.0,
                       kappa=-1):
        """
        Inputs:
            quad_params, dict with keys specified in quadrotor_params.py
            Q, the process noise covariance
            R, the measurement noise covariance
            x0, the initial filter state
            P0, the initial state covariance
            dt, the time between predictions
        """

        self.mass            = quad_params['mass'] # kg

        # Frame parameters
        self.c_Dx            = quad_params['c_Dx']  # drag coeff, N/(m/s)**2
        self.c_Dy            = quad_params['c_Dy']  # drag coeff, N/(m/s)**2
        self.c_Dz            = quad_params['c_Dz']  # drag coeff, N/(m/s)**2

        self.g = 9.81 # m/s^2

        # Filter parameters
        self.xhat = xhat0
        self.P = P0

        self.dt = dt

        self.N = self.xhat.shape[0]

        self.points = MerweScaledSigmaPoints(self.N, alpha=alpha, beta=beta, kappa=kappa)

        self.uk = np.array([self.g, 0, 0, 0])

        self.filter = UnscentedKalmanFilter(dim_x=self.N, dim_z=self.N, dt=dt, fx=self.f, hx=self.h, points=self.points)
        self.filter.x = xhat0
        self.filter.P = P0
        self.filter.R = R
        self.filter.Q = Q

        
    def step(self, ground_truth_state, controller_command, imu_measurement, mocap_measurement):
        """
        The step command will update the filter based on the following. 
        Inputs:
            ground_truth_state, the ground truth state is mainly there if it's necessary to compute certain portions of the state, e.g., actual thrust produced by the rotors. 
            controller_command, the controller command taken, this has to be converted to the appropriate control vector u depending on the filter model. 
            imu_measurement, contains measurements from an inertial measurement unit. These measurements are noisy, potentially biased, and potentially off-axis. The measurement
                        is specific acceleration, i.e., total force minus gravity. 
            mocap_measurement, provides noisy measurements of pose and twist. 

        Outputs:
            A dictionary with the following keys: 
                filter_state, containing the current filter estimate.
                covariance, containing the current covariance matrix.

        The ground truth state is supplied in case you would like to have "knowns" in your filter, or otherwise manipulate the state to create a custom measurement of your own desires. 
        IMU measurements are commonly used in filter setups, so we already supply these measurements as an input into the system. 
        Motion capture measurements are useful if you want noisy measurements of the pose and twist of the vehicle. 
        """

        # Extract the appropriate u vector based on the controller commands.
        self.uk = self.construct_control_vector(ground_truth_state, controller_command)

        # Construct the measurement vector yk
        orientation = Rotation.from_quat(copy.deepcopy(mocap_measurement['q']))
        euler_angles = orientation.as_euler('zyx', degrees=False)  # Get Euler angles from current orientation
        inertial_speed = mocap_measurement['v']
        body_speed = (orientation.as_matrix()).T@inertial_speed
        zk = np.array([euler_angles[0],                 # phi
                       euler_angles[1],                 # theta
                       euler_angles[2],                 # psi
                       body_speed[0],                   # vx
                       body_speed[1],                   # vy
                       body_speed[2],                   # vz
                       imu_measurement['accel'][0],     # body x acceleration
                       imu_measurement['accel'][1],     # body y acceleration
                       imu_measurement['accel'][2]      # body z acceleration
                       ])

        self.filter.predict()
        self.filter.update(zk)

        return {'filter_state': self.filter.x, 'covariance': self.filter.P}

    def f(self, xk, dt):
        """
        Process model
        """

        va = np.sqrt((xk[3]-xk[6])**2 + (xk[4]-xk[7])**2 + (xk[5]-xk[8])**2)  # Compute the norm of the airspeed vector

        # The process model is integrated using forward Euler. Below assumes Euler angles are given in order of [phi, theta, psi] (XYZ)
        # xdot = np.array([self.uk[1] + xk[0]*xk[1]*self.uk[2] + xk[2]*self.uk[3], 
        #                 self.uk[2] - xk[0]*self.uk[3], 
        #                 xk[0]*self.uk[2] + self.uk[3], 
        #                 -self.c_Dx/self.mass*(xk[3]-xk[6])*va + self.g*xk[1] + xk[4]*self.uk[3] - xk[5]*self.uk[1],
        #                 -self.c_Dy/self.mass*(xk[4]-xk[7])*va - self.g*xk[0] + xk[5]*self.uk[1] - xk[3]*self.uk[3], 
        #                 self.uk[0] - self.c_Dz/self.mass*(xk[5]-xk[8])*va - self.g + xk[3]*self.uk[2] - xk[4]*self.uk[1], 
        #                 0,
        #                 0,
        #                 0])

        # The process model, below assumes Euler angles are given in the order of [psi, theta, phi] (ZYX)
        xdot = np.array([xk[2]*self.uk[2] + self.uk[3], 
                        self.uk[2] - xk[2]*self.uk[3], 
                        self.uk[1] + xk[2]*xk[1]*self.uk[2] + xk[0]*self.uk[3], 
                        -self.c_Dx/self.mass*(xk[3]-xk[6])*va + self.g*xk[1] + xk[4]*self.uk[3] - xk[5]*self.uk[1],
                        -self.c_Dy/self.mass*(xk[4]-xk[7])*va - self.g*xk[2] + xk[5]*self.uk[1] - xk[3]*self.uk[3], 
                        self.uk[0] - self.c_Dz/self.mass*(xk[5]-xk[8])*va - self.g + xk[3]*self.uk[2] - xk[4]*self.uk[1], 
                        0,
                        0,
                        0])

        xkp1 = xk + xdot*dt

        return xkp1

    def h(self, xk):
        """
        Measurement model
        """

        h = np.zeros(xk.shape)

        va = np.sqrt((xk[3]-xk[6])**2 + (xk[4]-xk[7])**2 + (xk[5]-xk[8])**2)  # Compute the norm of the airspeed vector

        h[0:3] = np.hstack((np.eye(3), np.zeros((3,6))))@(xk)
        h[3:6] = np.hstack((np.zeros((3,3)), np.eye(3), np.zeros((3,3))))@(xk)

        h[6:] = np.array([-self.c_Dx/self.mass*(xk[3]-xk[6])*va, 
                          -self.c_Dy/self.mass*(xk[4]-xk[7])*va,
                          self.uk[0]-self.c_Dz/self.mass*(xk[5]-xk[8])*va])

        return h

    def construct_control_vector(self, ground_truth_state, controller_command):
        """
        Constructs control vector
        """
        uk = np.array([controller_command['cmd_thrust']/self.mass,    # Compute mass normalized thrust from the command thrust, note that this is not the actual thrust...
                       ground_truth_state['w'][0],                      # Body rate in x axis
                       ground_truth_state['w'][1],                      # Body rate in y axis
                       ground_truth_state['w'][2]]                      # Body rate in z axis
                       )

        return uk


import numpy as np
from scipy.spatial.transform import Rotation
import copy


class WindEKF:
    """
    Wind EKF: 
        Given approximate dynamics near level flight, the wind EKF will produce an estimate of the local wind vector acting on the body. 
        It requires knowledge of the effective drag coefficient on each axis, which would be determined either from real flight or computed in simulation, and the mass of the vehicle. 
        The inputs to the filter are the mass normalized collective thrust and the body rates on each axis. 
        Measurements of body velocity, Euler angles, and acceleration are provided to the vehicle. 
        The filter estimates the Euler angles, body velocities, and wind velocities. 

        State space: 
            xhat = [phi, theta, psi, u, v, w, windx, windy, windz]
        Input space: 
            u = [T/m, p, q, r]
    """

    def __init__(self, quad_params, Q=np.diag(np.concatenate([0.5*np.ones(3),0.7*np.ones(3),0.1*np.ones(3)])),
                                    R=np.diag(np.concatenate([0.0005*np.ones(3),0.0010*np.ones(3),np.sqrt(100/2)*(0.38**2)*np.ones(3)])), 
                                    xhat0=np.array([0,0,0, 0.1,0.05,0.02, 1.5,1.5,1.5]), 
                                    P0=1*np.eye(9),
                                    dt=1/100):
        """
        Inputs:
            quad_params, dict with keys specified in quadrotor_params.py
            Q, the process noise covariance
            R, the measurement noise covariance
            x0, the initial filter state
            P0, the initial state covariance
            dt, the time between predictions
        """
        # Quadrotor physical parameters.
        # Inertial parameters
        self.mass            = quad_params['mass'] # kg

        # Frame parameters
        self.c_Dx            = quad_params['c_Dx']  # drag coeff, N/(m/s)**2
        self.c_Dy            = quad_params['c_Dy']  # drag coeff, N/(m/s)**2
        self.c_Dz            = quad_params['c_Dz']  # drag coeff, N/(m/s)**2

        self.g = 9.81 # m/s^2

        # Filter parameters
        self.Q = Q
        self.R = R
        self.xhat = xhat0
        self.P = P0

        self.innovation = np.zeros((9,))

        self.dt = dt

        # Initialize the Jacobians at starting position and assuming hover thrust.  
        self.computeJacobians(self.xhat, np.array([self.g, 0, 0, 0]))

    def step(self, ground_truth_state, controller_command, imu_measurement, mocap_measurement):
        """
        This will perform both a propagate and update step in one for the sake of readability in other parts of the code.
        """
        self.propagate(ground_truth_state, controller_command)
        self.update(ground_truth_state, controller_command, imu_measurement, mocap_measurement)

        return self.pack_results()

    def propagate(self, ground_truth_state, controller_command):
        """
        Propagate occurs whenever an action u is taken.
        Inputs:
            ground_truth_state, the ground truth state is mainly there if it's necessary to compute certain portions of the state, e.g., actual thrust produced by the rotors. 
            controller_command, the controller command taken, this has to be converted to the appropriate control vector u depending on the filter model. 

        Outputs:
            xhat, the current state estimate after propagation
            P, the current covariance matrix after propagation
        """

        # Extract the appropriate u vector based on the controller commands.
        uk = self.construct_control_vector(ground_truth_state, controller_command)

        # First propagate the dynamics using the process model
        self.xhat = self.process_model(self.xhat, uk)

        # Update the covariance matrix using the linearized version of the dynamics
        self.computeJacobians(self.xhat, uk)
        self.P = self.A@self.P@(self.A.T) + self.Q

        return self.pack_results

    def update(self, ground_truth_state, controller_command, imu_measurement, mocap_measurement):
        """
        Update the estimate based on new sensor measurments. 
        Inputs:
            ground_truth_state, the ground truth state is mainly there if it's necessary to compute certain portions of the state, e.g., actual thrust produced by the rotors. 
            controller_command, the controller command taken, this has to be converted to the appropriate control vector u depending on the filter model. 
            imu_measurement, contains measurements from an inertial measurement unit. These measurements are noisy, potentially biased, and potentially off-axis. The measurement
                        is specific acceleration, i.e., total force minus gravity. 
            mocap_measurement, provides noisy measurements of pose and twist. 
        
        Outputs:
            xhat, the current state estimate after measurement update
            P, the current covariance matrix after measurement update
        """

        # Extract the appropriate u vector based on the controller commands.
        uk = self.construct_control_vector(ground_truth_state, controller_command)

        # Construct the measurement vector yk
        orientation = Rotation.from_quat(copy.deepcopy(mocap_measurement['q']))
        euler_angles = orientation.as_euler('zyx', degrees=False)  # Get Euler angles from current orientation
        inertial_speed = mocap_measurement['v']
        body_speed = (orientation.as_matrix()).T@inertial_speed
        yk = np.array([euler_angles[0],                 # phi
                       euler_angles[1],                 # theta
                       euler_angles[2],                 # psi
                       body_speed[0],                   # vx
                       body_speed[1],                   # vy
                       body_speed[2],                   # vz
                       imu_measurement['accel'][0],     # body x acceleration
                       imu_measurement['accel'][1],     # body y acceleration
                       imu_measurement['accel'][2]      # body z acceleration
                       ])
        
        # First linearize the measurement model. 
        self.computeJacobians(self.xhat, uk)

        # Now compute the Kalman gain
        K = self.P@(self.C.T)@np.linalg.inv(self.C@self.P@(self.C.T) + self.R)

        # Next compute the posteriori distribution
        self.innovation = yk - self.measurement_model(self.xhat, uk)
        self.xhat = self.xhat + K@self.innovation
        self.P = (np.eye(self.xhat.shape[0]) - K@self.C)@self.P

        return self.pack_results()

    def process_model(self, xk, uk):
        """
        Process model
        """

        va = np.sqrt((xk[3]-xk[6])**2 + (xk[4]-xk[7])**2 + (xk[5]-xk[8])**2)  # Compute the norm of the airspeed vector

        # The process model is integrated using forward Euler. 
        xdot = np.array([uk[1] + xk[0]*xk[1]*uk[2] + xk[2]*uk[3], 
                        uk[2] - xk[0]*uk[3], 
                        xk[0]*uk[2] + uk[3], 
                        -self.c_Dx/self.mass*(xk[3]-xk[6])*va + self.g*xk[1] + xk[4]*uk[3] - xk[5]*uk[1],
                        -self.c_Dy/self.mass*(xk[4]-xk[7])*va - self.g*xk[0] + xk[5]*uk[1] - xk[3]*uk[3], 
                        uk[0] - self.c_Dz/self.mass*(xk[5]-xk[8])*va - self.g + xk[3]*uk[2] - xk[4]*uk[1], 
                        0,
                        0,
                        0])

        xkp1 = xk + xdot*self.dt

        return xkp1
    
    def measurement_model(self, xk, uk):
        """
        Measurement model
        """

        h = np.zeros(xk.shape)

        va = np.sqrt((xk[3]-xk[6])**2 + (xk[4]-xk[7])**2 + (xk[5]-xk[8])**2)  # Compute the norm of the airspeed vector

        h[0:3] = np.hstack((np.eye(3), np.zeros((3,6))))@(xk)
        h[3:6] = np.hstack((np.zeros((3,3)), np.eye(3), np.zeros((3,3))))@(xk)

        h[6:] = np.array([-self.c_Dx/self.mass*(xk[3]-xk[6])*va, 
                          -self.c_Dy/self.mass*(xk[4]-xk[7])*va,
                          uk[0]-self.c_Dz/self.mass*(xk[5]-xk[8])*va])

        return h

    def computeJacobians(self, x, u):
        """
        Compute the Jacobians of the process and measurement model at the operating points x and u.
        """

        va = np.sqrt((x[3]-x[6])**2 + (x[4]-x[7])**2 + (x[5]-x[8])**2)  # Compute the norm of the airspeed vector

        # Partial derivatives of va for chain rule
        dvadu = (x[3]-x[6])/va
        dvadv = (x[4]-x[7])/va
        dvadw = (x[5]-x[8])/va
        dvadwx = -dvadu
        dvadwy = -dvadv
        dvadwz = -dvadw

        kx = self.c_Dx/self.mass
        ky = self.c_Dy/self.mass
        kz = self.c_Dz/self.mass

        vax = x[3] - x[6]
        vay = x[4] - x[7]
        vaz = x[5] - x[8]

        self.A = np.array([[x[1]*u[2], x[0]*u[2] + u[3], 0, 0, 0, 0, 0, 0, 0],
                           [-u[3], 0, 0, 0, 0, 0, 0, 0, 0],
                           [ u[2], 0, 0, 0, 0, 0, 0, 0, 0],
                           [0, self.g, 0, -kx*(dvadu*vax + va), u[3] - kx*(dvadv*vax), -u[1] - kx*(dvadw*vax), -kx*(dvadwx*vax-va), -kx*(dvadwy*vax), -kx*(dvadwz*vax)],
                           [-self.g, 0, 0, -ky*(dvadu*vay)-u[3], -ky*(dvadv*vay + va), u[1]-ky*(dvadw*vay), -ky*(dvadwx*vay), -ky*(dvadv*vay - va), -ky*(dvadw*vay)],
                           [0, 0, 0, u[2] - kz*(dvadu*vaz), -u[1] - kz*(dvadv*vaz), -kz*(dvadw*vaz + va), -kz*(dvadwx*vaz), -kz*(dvadwy*vaz), -kz*(dvadwz*vaz - va)],
                           [0, 0, 0, 0, 0, 0, 0, 0, 0],
                           [0, 0, 0, 0, 0, 0, 0, 0, 0],
                           [0, 0, 0, 0, 0, 0, 0, 0, 0]])

        self.C = np.array([[1, 0, 0, 0, 0, 0, 0, 0, 0],
                           [0, 1, 0, 0, 0, 0, 0, 0, 0],
                           [0, 0, 1, 0, 0, 0, 0, 0, 0],
                           [0, 0, 0, 1, 0, 0, 0, 0, 0],
                           [0, 0, 0, 0, 1, 0, 0, 0, 0],
                           [0, 0, 0, 0, 0, 1, 0, 0, 0],
                           [0, 0, 0, -kx*(dvadu*vax + va),  -kx*(dvadv*vax), -kx*(dvadw*vax), -kx*(dvadwx*vax-va), -kx*(dvadwy*vax), -kx*(dvadwz*vax)],
                           [0, 0, 0, -ky*(dvadu*vay), -ky*(dvadv*vay + va), -ky*(dvadw*vay), -ky*(dvadwx*vay), -ky*(dvadv*vay - va), -ky*(dvadw*vay)],
                           [0, 0, 0, -kz*(dvadu*vaz), -kz*(dvadv*vaz), -kz*(dvadw*vaz + va), -kz*(dvadwx*vaz), -kz*(dvadwy*vaz), -kz*(dvadwz*vaz - va)]])

        return

    def construct_control_vector(self, ground_truth_state, controller_command):
        """
        Constructs control vector
        """
        uk = np.array([controller_command['cmd_thrust']/self.mass,    # Compute mass normalized thrust from the command thrust, note that this is not the actual thrust...
                       ground_truth_state['w'][0],                      # Body rate in x axis
                       ground_truth_state['w'][1],                      # Body rate in y axis
                       ground_truth_state['w'][2]]                      # Body rate in z axis
                       )

        return uk

    def pack_results(self):
        # return {'euler_est': self.xhat[0:3], 'v_est': self.xhat[3:6], 'wind_est': self.xhat[6:9],
        #          'covariance': self.P, 'innovation': self.innovation}
        return {'filter_state': self.xhat, 'covariance': self.P}


"""
Physical parameters for the AscTec Hummingbird. Values parameterize the inertia, motor dynamics, 
rotor aerodynamics, parasitic drag, and rotor placement. 
Additional sources:
    https://digitalrepository.unm.edu/cgi/viewcontent.cgi?article=1189&context=ece_etds
    https://ieeexplore.ieee.org/stamp/stamp.jsp?arnumber=7991501
Notes:
    The k_d and k_z terms are an order of magnitude smaller because 10^-3 was too large. 
"""
import numpy as np

d = 0.17 # Arm length

quad_params = {

    # Inertial properties
    'mass': 0.500,       # kg
    'Ixx':  3.65e-3,    # kg*m^2
    'Iyy':  3.68e-3,    # kg*m^2
    'Izz':  7.03e-3,    # kg*m^2
    'Ixy':  0.0,        # kg*m^2
    'Iyz':  0.0,        # kg*m^2 
    'Ixz':  0.0,        # kg*m^2

    # Geometric properties, all vectors are relative to the center of mass.
    'num_rotors': 4,                        # for looping over each actuator
    'rotor_radius': 0.10,                   # rotor radius, in meters
    'rotor_pos': {  
                    'r1': d*np.array([ 0.70710678118, 0.70710678118, 0]),    # Location of Rotor 1, meters
                    'r2': d*np.array([ 0.70710678118,-0.70710678118, 0]),    # Location of Rotor 2, meters
                    'r3': d*np.array([-0.70710678118,-0.70710678118, 0]),    # Location of Rotor 3, meters
                    'r4': d*np.array([-0.70710678118, 0.70710678118, 0]),    # Location of Rotor 4, meters
                 },

    'rotor_directions': np.array([1,-1,1,-1]),  # This dictates the direction of the torque for each motor. 

    'rI': np.array([0,0,0]), # location of the IMU sensor, meters

    # Frame aerodynamic properties
    'c_Dx': 0.5e-2,  # parasitic drag in body x axis, N/(m/s)**2
    'c_Dy': 0.5e-2,  # parasitic drag in body y axis, N/(m/s)**2
    'c_Dz': 1e-2,  # parasitic drag in body z axis, N/(m/s)**2

    # Rotor properties
    'k_eta': 5.57e-06,          # thrust coefficient N/(rad/s)**2
    'k_m':   1.36e-07,          # yaw moment coefficient Nm/(rad/s)**2
    'k_d':   1.19e-04,          # rotor drag coefficient N/(rad*m/s**2) = kg/rad
    'k_z':   2.32e-04,          # induced inflow coefficient N/(rad*m/s**2) = kg/rad
    'k_flap': 0.0,              # Flapping moment coefficient Nm/(rad*m/s**2) = kg*m/rad

    # Motor properties
    'tau_m': 0.005,             # motor response time, seconds
    'rotor_speed_min': 0,       # rad/s
    'rotor_speed_max': 1500,    # rad/s
    'motor_noise_std': 50,      # rad/s

}


"""
Utility scripts for the Dryden wind model
The Dryden Gust model is implemented using this package: 
https://github.com/goromal/wind-dynamics

"""
import numpy as np

class GustModelBase():
    """
    """

    def __init__(self, V, L, sigma, dt=0.05):
        """
        Inputs: 
            V, average velocity through medium, m/s
            L, characteristic length scale, m
            sigma, turbulence intensity
        """

        self.dt = dt
        b = 2*np.sqrt(3)*L/V
        c = 2*L/V

        self.alpha = sigma*np.sqrt(2*L/np.pi/V)
        self.beta = self.alpha*b
        self.delta = 2*c
        self.gamma = c*c

        self.u_km1 = 0
        self.u_km2 = 0
        self.y_km1 = 0
        self.y_km2 = 0

        self.initialized = True
        return 
    
    def run(self, dt):
        """
        """

        C1 = 1.0 + 2*self.delta/dt + 4*self.gamma/dt/dt
        C2 = 2.0 - 8*self.gamma/dt/dt
        C3 = 1.0 - 2*self.delta/dt + 4*self.gamma/dt/dt
        C4 = self.alpha + 2*self.beta/dt
        C5 = 2*self.alpha
        C6 = self.alpha - 2*self.beta/dt

        u_k = np.random.uniform(-1,1)
        y_k = (C4*u_k + C5*self.u_km1 + C6*self.u_km2 - C2*self.y_km1 - C3*self.y_km2)/C1

        self.u_km2 = self.u_km1
        self.u_km1 = u_k
        self.y_km2 = self.y_km1
        self.y_km1 = y_k

        return y_k
    
    def integrate(self, dt):
        """
        """

        if (dt > self.dt):
            t = 0
            y = 0
            while (t < dt):
                t_inc = min(self.dt, dt - t)
                y = self.run(t_inc)
                t += t_inc
            return y
        else:
            return self.run(dt)
        
class DrydenWind():
    """
    """

    def __init__(self, wx_nominal, wy_nominal, wz_nominal, 
                       wx_sigma, wy_sigma, wz_sigma, 
                       altitude=2.0):
        
        self.wx_nominal, self.wy_nominal, self.wz_nominal = wx_nominal, wy_nominal, wz_nominal
        self.altitude = altitude

        Lz_ft = 3.281 * altitude
        Lx_ft = Lz_ft / ((0.177 + 0.000823 * Lz_ft)**(1.2))
        Ly_ft = Lx_ft

        self.wx_gust = GustModelBase(1.0, Lx_ft/3.281, wx_sigma)
        self.wy_gust = GustModelBase(1.0, Ly_ft/3.281, wy_sigma)
        self.wz_gust = GustModelBase(1.0, Lz_ft/3.281, wz_sigma)

        self.initialized = True

        return
    
    def getWind(self, dt):
        """
        """

        if self.initialized:
            wind_vector = np.array([self.wx_nominal, self.wy_nominal, self.wz_nominal]) + np.array([self.wx_gust.integrate(dt), self.wy_gust.integrate(dt), self.wz_gust.integrate(dt)])
        else:
            wind_vector = np.array([0,0,0])

        return wind_vector
    

if __name__=="__main__":

    import matplotlib.pyplot as plt
    import numpy as np

    t = 100.0
    dt = 0.05
    x_mean = 0.0
    y_mean = 1.0
    z_mean = 0.0
    x_sigma = 10.0
    y_sigma = 0.0
    z_sigma = 70.0

    plt.rcParams.update({
        "text.usetex": True,
        "font.family": "serif",
        "font.serif": ["Palatino"],
    })

    wind = DrydenWind(x_mean, y_mean, z_mean, x_sigma, y_sigma, z_sigma)

    n = int(np.floor(t / dt)) + 1
    x = np.linspace(0, t, n)
    y = np.zeros((3,x.size))

    for i in range(n):
        y[:,i] = wind.getWind(dt)

    fig, axs = plt.subplots(3)
    fig.suptitle('Dryden Gust Velocities\n$\sigma_x=%f$, $\sigma_y=%f$, $\sigma_z=%f$' % (x_sigma, y_sigma, z_sigma))
    axs[0].plot([0, t],[x_mean, x_mean])
    axs[0].plot(x, y[0,:])
    axs[0].set_ylabel('X Velocity')
    axs[1].plot([0, t],[y_mean, y_mean])
    axs[1].plot(x, y[1,:])
    axs[1].set_ylabel('Y Velocity')
    axs[2].plot([0, t],[z_mean, z_mean])
    axs[2].plot(x, y[2,:])
    axs[2].set_ylabel('Z Velocity')
    axs[2].set_xlabel('Time (s)')

    plt.show()


import numpy as np
from scipy.spatial.transform import Rotation
import matplotlib.pyplot as plt
import time as clk

from rotorpy.simulate import simulate
from rotorpy.utils.plotter import *
from rotorpy.world import World

from rotorpy.utils.postprocessing import unpack_sim_data

import os


class Environment():
    """
    Sandbox represents an instance of the simulation environment containing a unique vehicle, 
    controller, trajectory generator, wind profile. 

    """

    def __init__(self, vehicle,                 # vehicle object, must be specified. 
                       controller,              # controller object, must be specified.
                       trajectory,              # trajectory object, must be specified.
                       wind_profile = None,     # wind profile object, if none is supplied it will choose no wind. 
                       imu = None,              # imu sensor object, if none is supplied it will choose a default IMU sensor.
                       mocap = None,            # mocap sensor object, if none is supplied it will choose a default mocap.
                       world        = None,     # The world object
                       estimator    = None,     # estimator object
                       sim_rate     = 100,      # The update frequency of the simulator in Hz
                       safety_margin = 0.25,    # The radius of the safety region around the robot. 
                       ):

        self.sim_rate = sim_rate
        self.vehicle = vehicle
        self.controller = controller
        self.trajectory = trajectory

        self.safety_margin = safety_margin

        if world is None:
            # If no world is specified, assume that it means that the intended world is free space.
            wbound = 3 
            self.world = World.empty((-wbound, wbound, -wbound, 
                                       wbound, -wbound, wbound))
        else:
            self.world = world

        if wind_profile is None:
            # If wind is not specified, default to no wind. 
            from rotorpy.wind.default_winds import NoWind
            self.wind_profile = NoWind()
        else:
            self.wind_profile = wind_profile
        
        if imu is None:
            # In the event of specified IMU, default to 0 bias with white noise with default parameters as specified below. 
            from rotorpy.sensors.imu import Imu
            self.imu = Imu(p_BS = np.zeros(3,),
                           R_BS = np.eye(3),
                           sampling_rate=sim_rate)
        else:
            self.imu = imu

        if mocap is None:
            # If no mocap is specified, set a default mocap. 
            # Default motion capture properties. Pretty much made up based on qualitative comparison with real data from Vicon. 
            mocap_params = {'pos_noise_density': 0.0005*np.ones((3,)),  # noise density for position 
                    'vel_noise_density': 0.0010*np.ones((3,)),          # noise density for velocity
                    'att_noise_density': 0.0005*np.ones((3,)),          # noise density for attitude 
                    'rate_noise_density': 0.0005*np.ones((3,)),         # noise density for body rates
                    'vel_artifact_max': 5,                              # maximum magnitude of the artifact in velocity (m/s)
                    'vel_artifact_prob': 0.001,                         # probability that an artifact will occur for a given velocity measurement
                    'rate_artifact_max': 1,                             # maximum magnitude of the artifact in body rates (rad/s)
                    'rate_artifact_prob': 0.0002                        # probability that an artifact will occur for a given rate measurement
            }
            from rotorpy.sensors.external_mocap import MotionCapture
            self.mocap = MotionCapture(sampling_rate=sim_rate, mocap_params=mocap_params, with_artifacts=False)
        else:
            self.mocap = mocap

        if estimator is None:
            # In the likely case where an estimator is not supplied, default to the null state estimator. 
            from rotorpy.estimators.nullestimator import NullEstimator
            self.estimator = NullEstimator()
        else:
            self.estimator = estimator

        return 

    def run(self,   t_final      = 10,       # The maximum duration of the environment in seconds
                    use_mocap    = False,    # boolean determines if the controller should use
                    terminate    = False,
                    plot            = False,    # Boolean: plots the vehicle states and commands   
                    plot_mocap      = True,     # Boolean: plots the motion capture pose and twist measurements
                    plot_estimator  = True,     # Boolean: plots the estimator filter states and covariance diagonal elements
                    plot_imu        = True,     # Boolean: plots the IMU measurements
                    animate_bool    = False,    # Boolean: determines if the animation of vehicle state will play. 
                    animate_wind    = False,    # Boolean: determines if the animation will include a wind vector.
                    verbose         = False,    # Boolean: will print statistics regarding the simulation. 
                    fname   = None      # Filename is specified if you want to save the animation. Default location is the home directory. 
                    ):

        """
        Run the simulator
        """

        self.t_step = 1/self.sim_rate
        self.t_final = t_final
        self.t_final = t_final
        self.terminate = terminate
        self.use_mocap = use_mocap

        start_time = clk.time()
        (time, state, control, flat, imu_measurements, imu_gt, mocap_measurements, state_estimate, exit) = simulate(self.world,
                                                                                                                    self.vehicle.initial_state,
                                                                                                                    self.vehicle,
                                                                                                                    self.controller,
                                                                                                                    self.trajectory,
                                                                                                                    self.wind_profile,
                                                                                                                    self.imu,
                                                                                                                    self.mocap,
                                                                                                                    self.estimator,
                                                                                                                    self.t_final,
                                                                                                                    self.t_step,
                                                                                                                    self.safety_margin,
                                                                                                                    self.use_mocap,
                                                                                                                    terminate=self.terminate,
                                                                                                                    )
        if verbose:
            # Print relevant statistics or simulator status indicators here
            print('-------------------RESULTS-----------------------')
            print('SIM TIME -- %3.2f seconds | WALL TIME -- %3.2f seconds' % (min(self.t_final, time[-1]) , (clk.time()-start_time)))
            print('EXIT STATUS -- '+exit.value)

        self.result = dict(time=time, state=state, control=control, flat=flat, imu_measurements=imu_measurements, imu_gt=imu_gt, mocap_measurements=mocap_measurements, state_estimate=state_estimate, exit=exit)

        visualizer = Plotter(self.result, self.world)

        # Remove gif or mp4 in filename if it exists (the respective functions will add appropriate extensions)
        if fname is not None:
            if ".gif" in fname:
                fname = fname.replace(".gif", "")
            if ".mp4" in fname:
                fname = fname.replace(".mp4", "")
                
        if animate_bool:
            # Do animation here
            visualizer.animate_results(fname=fname, animate_wind=animate_wind)
        if plot:
            # Do plotting here
            visualizer.plot_results(fname=fname,plot_mocap=plot_mocap,plot_estimator=plot_estimator,plot_imu=plot_imu)
            if not animate_bool:
                plt.show()

        return self.result
    
    def save_to_csv(self, fname):
        """
        Save the simulation data in self.results to a file. 
        """

        if self.result is None:
            print("Error: cannot save if no results have been generated! Aborting save.")
            return
        else:
            if not ".csv" in fname:
                fname = fname + ".csv"
            path = os.path.join(os.path.dirname(__file__),'data_out',fname)
            dataframe = unpack_sim_data(self.result)
            dataframe.to_csv(path)


if __name__=="__main__":
    from rotorpy.vehicles.crazyflie_params import quad_params
    from rotorpy.trajectories.hover_traj import HoverTraj

    from rotorpy.vehicles.multirotor import Multirotor
    from rotorpy.controllers.quadrotor_control import SE3Control

    sim = Environment(vehicle=Multirotor(quad_params),
                      controller=SE3Control(quad_params),
                      trajectory=HoverTraj(),
                      sim_rate=100
                      )
    
    result = sim.run(t_final=1, plot=True)